"""CCAM DRS post-processing"""
import os
import re
import argparse
from pathlib import Path
import configparser as cp
import xarray as xr
import glob
import sys
import axiom.utilities as au
import axiom.schemas as axs
import numpy as np
import importlib
from pprint import pprint
from calendar import monthrange
import pandas as pd
from datetime import datetime, timedelta
from uuid import uuid4
import re
from axiom.exceptions import ResolutionDetectionException, MalformedDRSJSONPayloadException
from cerberus import Validator
from axiom.drs.domain import Domain
from axiom.config import load_config
import shutil


def is_fixed_variable(config, variable):
    """Determine if a variable is listed in in the config as fixed.

    Args:
        config (dict): Configuration dictionary
        variable (str): Variable name.

    Returns:
        bool: True if fixed, False otherwise.
    """
    return variable in config['fixed_variables']


def get_template(config, key):
    """Get an interpolation template out of the config.

    Args:
        config (dict): Dictionary.
        key (str): Template key.

    Returns:
        str: Template.
    """
    # Get the template out of the configuration
    template = config['templates'][key]

    # Convert list to string (helps with really long templates)
    if isinstance(template, list):

        # The delimiter is the first item in the list
        template = template[0].join(template[1:])

    return template


def get_domains(resolution, frequency, variable_fixed, no_frequencies):
    """Get the domains for the arguments provided.

    Args:
        resolution (int): Resolution
        frequency (str): Frequency
        variable_fixed (bool): Is the variable fixed?
        no_frequencies (bool): True if no frequencies were provided.

    Raises:
        Exception: When the arguments provided don't yield any domain information.

    Returns:
        dict: Dictionary of domain information.
    """

    is_cordex = frequency == 'cordex'

    domains, frequencies, resolution_dir, degrees = None, None, None, None

    # Quickly bail if necessary
    assert resolution in [50, 5, 10, 12]

    if resolution == 50:

        degrees = 0.5
        resolution_dir = f'{resolution}km'

        if not is_cordex:
            domains = ['AUS-50']

        else:

            domains = ['AUS-44i']
            frequencies = ['1D', '1M']

            if variable_fixed:
                frequencies = ['1D']

        if no_frequencies:
            frequencies = ['3H', '1D', '1M']

    elif resolution == 5:

        domains = ['VIC-5']
        degrees = 0.05
        resolution_dir = f'{resolution}km'

    elif resolution == 10:

        domains = ['SEA-10', 'TAS-10']
        resolution_dir = '5km',
        degrees = 0.1

    elif resolution == 12:

        if not is_cordex:
            raise Exception('Domain not known.')

        domains = ['AUS-44i']
        frequencies = ['1D', '1M']

        if variable_fixed:
            frequencies = ['1D']

    # Return everything
    return dict(
        domains=domains,
        frequencies=frequencies,
        resolution_dir=resolution_dir,
        degrees=degrees
    )


def get_meta(config, meta_key, key):
    """Load metadata on top of a _default key, if applicable.

    Args:
        config (dict): Configuration dictionary.
        meta_key (str): Top-level key.
        key (str): Key.

    Returns:
        dict: Metadata dictionary, with key loaded over _default.
    """

    # Start with nothing
    _meta = dict()

    # Load a default, if it exists
    if '_default' in config[meta_key].keys():
        _meta.update(config[meta_key]['_default'])

    # Load what was requested over top, if it exists
    if key in config[meta_key].keys():
        _meta.update(config[meta_key][key])
    else:
        au.get_logger(__name__).warn(f'Key {key} is not present in {meta_key}, loading defaults only.')

    return _meta


def metadata(obj, **kwargs):
    """Add metadata to an xarray object.

    Args:
        da (xarray.DataArray or xarray.Dataset): xarray object.

    Returns:
        xarray.DataArray or xarray.Dataset : Same as caller
    """
    for key,value in kwargs.items():
        obj.attrs[key] = value

    return obj


def standardise_units(ds):
    """Standardise units.

    Only converts mm to m for now.

    Args:
        ds (xarray.Dataset): Dataset

    Returns:
        xarray.Dataset: Dataset with units standardised.
    """

    for variable in ds.data_vars.keys():
        da = ds[variable]

        # Convert mm to m
        if ds[variable].attrs['units'] == 'mm':
            ds[variable] = metadata(da / 1000, units='m')

    return ds


def detect_resolution(paths):
    """Attempt to detect the input resolution from a list of paths.

    Args:
        paths (list): List of file paths.

    Raises:
        ResolutionDetectionException: If there are too many possible resolutions in a path.
        ResolutionDetectionException: If there are inconsistent resolutions detected between paths.

    Returns:
        int: Resolution in km.
    """

    # Set up pattern for searching
    regex = r'([0-9.]*km)'
    res = None

    for path in paths:

        matches = re.findall(regex,  path)

        # Too many  options to choose from
        if len(matches) != 1:
            raise ResolutionDetectionException(f'Unable to detect resolution from {path}, there are too many possibilities.')

        # First detection
        if res is None:
            res = matches[0]

        # Already detected, but not the same
        elif res != matches[0]:
            raise ResolutionDetectionException(f'Detected resolutions are inconsistent between supplied paths.')

    # Made it this far, we have a detectable resolution
    return float(res.replace('km', ''))


def input_files_exist(paths):
    """Ensure all the input files actually exist.

    Args:
        paths (list): List of paths.

    Returns:
        bool: True if they all exist, False otherwise.
    """

    for path in paths:
        if not os.path.isfile(path):
            return False

    return True


def _center_date(dt):
    """Centre the date for compatibility with CDO-processed data.

    Args:
        dt (object): Date object.

    Returns:
        same as called: Date object with day set to middle of the month.
    """
    num_days = monthrange(dt.year, dt.month)[1]
    return dt.replace(day=num_days // 2)


def postprocess_cordex(ds):
    """For CORDEX processing, there is some minor postprocessing that happens.

    Args:
        ds (xarray.Dataset): Data.

    Returns:
        xarray.Dataset: Data with postprocessing applied.
    """

    # Time coordinates need to be centered into the middle of the month
    centered_times = ds.time.to_pandas().apply(_center_date).values
    ds = ds.assign_coords(dict(time=centered_times))

    return ds


def parse_domain_directive(directive):
    """Parse a domain directive (i.e. from CLI).

    Args:
        directive (str): Domain directive of the form name,dx,lat_min,lat_max,lon_min,lon_max
    
    Returns:
        dict : Domain dictionary
    
    Raises:
        ValueError : If the domain cannot be parsed.
    """
    name, *directives = directive.split(',')
    dx, lat_min, lat_max, lon_min, lon_max = map(float, directives)
    return dict(
        name=name,
        dx=dx,
        lat_min=lat_min,
        lat_max=lat_max,
        lon_min=lon_min,
        lon_max=lon_max
    )


def parse_domain(directive):
    """Parse a domain directive.

    Domains are of the form: "name,fx,lat_min,lat_max,lon_min,lon_max"

    Args:
        directive (str) : Domain directive of the form name,dx,lat_min,lat_max,lon_min,lon_max.

    Returns:
        dict : Domain specification.

    Raises:
        AssertionError : When the directive is missing componenents.
        TypeError : When the directive is unable to be parsed.
    """
    segments = directive.split(',')
    assert len(segments) == 6
    name = segments[0]
    dx, lat_min, lat_max, lon_min, lon_max = [float(s) for s in segments[1:]]

    # Return a dictionary of the parsed information.
    return dict(
        name=name,
        dx=dx,
        lat_min=lat_min,
        lat_max=lat_max,
        lon_min=lon_min,
        lon_max=lon_max
    )


def load_domain_config():
    """Load domain configuration out of installed data dir.

    Returns:
        configparser.Config: Configuration object.
    """
    return load_config('domains')


def get_domain(key):
    """Load a domain out of the internal configuration.

    Args:
        key (str): Domain key or parseable domain directive.
    
    Returns:
        axiom.domain.Domain : Domain object.
    """

    # Load the domain configuration as a configuration object.
    domain_config = load_domain_config()
    return Domain.from_config(key, domain_config)


def is_registered_domain(key):
    """Quick shortcut to see if the domain is already registered in the system.

    Args:
        key (str): Domain key.

    Returns:
        bool: True if registered, False otherwise.
    """
    return key in load_domain_config().keys()


def load_processor(model_key, proc_type='pre'):
    """Load a pre-or-post processor for the model, if one exists.

    Args:
        model_key (str): Model.

    Returns:
        callable: Function that takes an xarray.Dataset as input.
    """
    logger = au.get_logger(__name__)

    try:
        mod = importlib.import_module(f'axiom.drs.processing.{model_key}')
        processor = getattr(mod, f'{proc_type}process_{model_key}')
        logger.info(f'Found {proc_type}processor for {model_key}')
        return processor
    except:
        logger.warning(f'No {proc_type}processor found for {model_key}, returning empty function.')
        return lambda ds, *args, **kwargs: ds

def load_preprocessor(model_key):
    """Shorthand for the load_processor function.

    Args:
        model_key (str): Model
    """
    return load_processor(model_key=model_key, proc_type='pre')


def load_postprocessor(model_key):
    """Shorthand for the load_processor function.

    Args:
        model_key (str): Model
    """
    return load_processor(model_key=model_key, proc_type='post')



def interpolate_context(context):
    """Interpolate the context dictionary into itself, filling all placeholders.

    Args:
        context (dict): Context dictionary.
    
    Returns:
        dict : Interpolated context.
    """
    logger = au.get_logger(__name__)
    for key, value in context.items():
        # context[key] = value % context
        new_value = str(value) % context
        logger.debug(f'{key} = {new_value}')
        context[key] = new_value
    
    return context


def get_uninterpolated_placeholders(string):
    """Check if a string has any remaining uninterpolated values.

    Args:
        string (string): String object.
    
    Returns:
        list : List of uninterpolated values.
    """
    # Regex to find matches
    matches = re.findall(r'%\(([a-zA-Z0-9_-]+)\)s', string)

    # Convert to set to remove duplicates, convert back and return
    return sorted(list(set(matches)))


def detect_input_frequency(ds):
    """Detect the (closest) input frequency of the data.

    Args:
        ds (xarray.Dataset): Input data.
    
    Returns:
        str : Pandas offset directive.
    """

    # Bail out if time-invariant
    if is_time_invariant(ds):
        return 'fx'

    # Take the difference between the first two time steps
    diff = (ds.time.data[1] - ds.time.data[0])
    total_seconds =  pd.to_timedelta([diff]).total_seconds()

    # TODO: These should be configurable
    intervals = '1H,3H,6H,1D,1M'.split(',')
    seconds = [
        1*60*60,
        3*60*60,
        6*60*60,
        24*60*60,
        28*24*60*60
    ]

    # Take the closest
    ix = np.argmin(np.abs(np.array(seconds) - total_seconds))
    return intervals[ix]


def is_time_invariant(ds):
    """Test if the dataset is time-invariant (has no time coordinate)

    Args:
        ds (xarray.Dataset or xarray.DataArray): Data
    
    Return:
        bool : True if no 'time' coordinate detected, False otherwise
    """
    return 'time' not in list(ds.coords.keys())


def is_error_recoverable(exception, recoverable_errors):
    """Determine if an error is recoverable based on the presence of certain text in the stack trace.

    Args:
        exception (Exception): Exception object, with message attribute.
        recoverable_errors (list): List of recoverable errors (regexes) that are permitted.
    
    Returns:
        bool : True if recoverable, False otherwise.
    """
    if hasattr(exception, 'message'):
        message = exception.message
    else:
        message = str(exception)

    # Check for the pattern
    for pattern in recoverable_errors:
        if re.search(pattern, message):
            return True
    
    return False


def assemble_qsub_vars(**kwargs):
    """Assemble variables into a qsub-compliant -v format, without the -v.

    Args:
        **kwargs (dict): Keyword arguments to convert into qsub variables.
    
    Returns:
        str : String of qsub variables.
    """
    return ','.join([f'{k}={v}' for k, v in kwargs.items()])


def assemble_qsub_command(jobscript, directives, **context):
    """Assemble the qsub command.
    
    Args:
        jobscript (str) : Path to the jobscript.
        directives (list) : List of directives to interpolate.
        **context (dict) : Context dictionary to interpolate into the directives.
    
    Returns:
        str : The qsub command.
    """
    # Collapse the directives into a single string and interpolate
    directives = ' '.join(directives) % context
    return f'qsub {directives} {jobscript}'

    
def generate_user_config():
    """Generate the user .axiom directory with all installed data files.
    
    Args:
        overwrite (bool): Overwrite existing .axiom directory.
    """

    logger = au.get_logger(__name__)

    # Work out the users home directory (os independent)
    axiom_dir = os.path.join(
        str(Path.home()),
        '.axiom'
    )

    # Does it already exist? back it up with a timestamp of the current tim
    if os.path.exists(axiom_dir):
        timestamp = datetime.now().strftime('%Y%m%d%H%M%S')
        backup_dir = f'{axiom_dir}_{timestamp}'
        print(f'Backing up existing .axiom directory to {backup_dir}')
        shutil.move(axiom_dir, backup_dir)

    # Create the directory and copy the files over.
    logger.info(f'Creating new .axiom directory at {axiom_dir}')
    os.makedirs(axiom_dir, exist_ok=True)
    
    config_files = au.auto_glob(os.path.join(
        importlib.resources.files('axiom'),
        'data/*.json'
    ))

    for src in config_files:
        dst = os.path.join(axiom_dir, os.path.basename(src))
        logger.info(f'Copying {src} to {dst}')
        shutil.copy(src, dst)
    
    logger.info('User configuration generated successfully.')


def filter_by_variable_name(filepaths, variable):
    """Filter for filenames that include the variable name.

    Args:
        filepaths (list): List of filepaths.

    Returns:
        list : List of filepaths that include the variable name.
    """
    config = load_config('drs')

    # Bail out if not filtering
    if not config['filename_filtering']['variable']:
        return filepaths
    
    # Get the regex information, interpolate the variable name
    pattern_template = config['filename_filtering']['variable_regex']
    pattern = pattern_template % dict(variable=variable)

    # Filter for the variable name
    filtered_filepaths = list()
    for filepath in filepaths:
        filename = os.path.basename(filepath)
        if re.search(pattern, filename):
            filtered_filepaths.append(filepath)
    
    return filtered_filepaths


def get_start_and_end_dates(year, output_frequency):
    """Get the start and end dates for interpolation context.

    Args:
        year (int): Current year
        output_frequency (str): Frequency.
    """
    start_date = f'{year}0101' if output_frequency[-1] != 'M' else f'{year}01'
    end_date = f'{year}1231' if output_frequency[-1] != 'M' else f'{year}12'
    return start_date, end_date


def generate_years_list(start_year, end_year):
    """Generate a list of years (decades) to process.

    Args:
        start_year (int): Start year.
        end_year (int): End year.

    Returns:
        iterator : Years to process.
    """
    return range(start_year, end_year+1, 10)



"""Conversion routines to convert between units automatically. EXPERIMENTAL."""
from axiom.decorators import metadata


@metadata(units='m')
def mm2m(da):
    """Convert mm to m.

    Args:
        da (xarray.DataArray): Data in mm.

    Returns:
        xarray.DataArray: Data in m.
    """
    return da / 1000.0

conversion_mappings = {
    'mm:m': mm2m
}

def ensure_units(da, expected_units):
    """Ensure that a given DataArray has the correct units, converting if needed.

    Args:
        da (xarray.DataArray): Data.
        expected_units (ste): Units value.

    Returns:
        xarray.DataArray: Data with units converted.
    
    Raises:
        KeyError : When no unit attribute is present.
        KeyError : When no converter is registered between the old and new units.
        Exception: When the converter fails to convert the units.
    """
    
    # Get the current units information
    current_units = da.attrs['units']
    
    # No conversion required
    if current_units == expected_units:
        return da

    # Lookup the conversion routine
    converter = conversion_mappings[f'{current_units}:{expected_units}']

    # Call it
    return converter(da)
    


"""Quality-Assurance module."""
import glob
import pandas as pd
import numpy as np
import axiom.utilities as au
from tqdm import tqdm
import os


def check_timeseries_variable(search_template, variable, start_year, end_year):
    """Check the timeseries for a given variable.

    Args:
        search_template (str): Globbable search path with %(variable)s and %(year)s placeholders.
        variable (str or list): Variable name(s).
        start_year (int): Start year.
        end_year (int): End year.
    
    Returns:
        pandas.DataFrame: DataFrame of results.
    """

    rows = list()

    for _variable in tqdm(au.pluralise(variable)):

        for year in range(start_year, end_year + 1):

            context = dict(variable=_variable, year=year)

            filepath_search = search_template % context
            # filepath_search = au.interpolate_template(search_template, **context)
            filepaths = sorted(glob.glob(filepath_search))

            # row = dict(filepath_search=filepath_search)
            row = dict(variable=_variable, year=year, filepath_search=filepath_search)

            # Success: file is present
            if len(filepaths) == 1:

                # Get the filesize in mb
                num_bytes = os.path.getsize(filepaths[0])
                filesize_mb = num_bytes / (1024 * 1024)
                _row = dict(status='SUCCESS', comment='File is present.', size=filesize_mb)

            # Error: no file
            elif len(filepaths) == 0:
                _row = dict(status='ERROR', comment='File is missing.', size=np.nan)

            # Error: multiple files detected
            else:
                _row = dict(status='ERROR', comment='Multiple files detected.', size=np.nan)

            row.update(_row)
            rows.append(row)
    
    df = pd.DataFrame.from_records(rows)
    return df


"""Reporting functions."""
from tabulate import tabulate
from axiom import __version__


def _is_missing(e):
    """Check if error was for a missing attribute.

    Args:
        e (cerberus.errors.ValidationError): Error object.

    Returns:
        bool: True/False
    """
    return isinstance(e.constraint, bool) and e.constraint == True and e.value == None


def _is_incorrect_type(e):
    """Check if error was for an incorrect attribute data type.

    Args:
        e (cerberus.errors.ValidationError): Error object.

    Returns:
        bool: True/False
    """
    return e.schema_path[-1] == 'type'


def _is_incorrect_value(e):
    """Check if error was for an incorrect attribute value.

    Args:
        e (cerberus.errors.ValidationError): Error object.

    Returns:
        bool: True/False
    """
    return e.schema_path[-1] == 'allowed'


def generate_report(validator, input_filepath, report_filepath):
    """Generate a report for a validator.

    Args:
        validator (axiom.validation.Validator): Validator object.
        input_filepath (str) : Path to  the file that was validated.
        report_filepath (str): Path to which to write the report.
    """

    # Create the headings and lines array
    table_heading = ['Variable', 'Attribute', 'Error']
    table_lines = list()

    # Global errors
    for e in validator.errors['_global']:

        variable = '_global'
        attribute = e.document_path[0]

        # Attribute missing
        if _is_missing(e):
            error = 'Attribute is missing'

        # Incorrect type
        elif _is_incorrect_type(e):
            expected_type = e.constraint
            error = f'Incorrect type, should be {expected_type}'

        else:
            error = 'Unknown error'

        table_lines.append([variable, attribute, error])


    # Errors per variable
    for key, errors in validator.errors['variables'].items():

        variable = key

        for e in errors:

            attribute = e.document_path[0]

            # Missing
            if _is_missing(e):
                error = "Attribute is missing."

            # Incorrect type
            elif _is_incorrect_type(e):
                expected_type = e.constraint
                error = f'Incorrect type, should be {expected_type}'

            # Incorrect value
            elif _is_incorrect_value(e):
                allowed_values = ','.join(e.constraint)
                error = f'Incorrect value, must be one of {allowed_values}'

            else:
                error = 'Unknown error'

            line = [variable, attribute, error]
            table_lines.append(line)


    # Tabulate reults
    table = tabulate(table_lines, table_heading, tablefmt='grid')

    # Determine the status
    status = 'PASSED' if validator.is_valid else 'FAILED'

    # Generate the heading.
    lines = list()
    lines.append(f'Axiom Validator {__version__}')
    lines.append(f'Report generated: {validator.date_validated}')
    lines.append(f'schema_filepath: {validator.schema_filepath}')
    lines.append(f'input_filepath: {input_filepath}')
    lines.append(f'Status: {status}\n')

    # Put it all together and dump to file
    content = '\n'.join(lines)
    content = content + table + '\n'
    open(report_filepath, 'w').write(content)



"""Test the configuration object."""
from axiom.config import Config, load_config


def test_boolean_fallback():
    """Test that a missing config key returns False."""
    c = Config()
    assert c.missing_key == False


def test_keys_in_installed_drs_json():
    """A test to ensure that new local dev keys have been added to the data/drs.json file."""

    user_config = load_config('drs', defaults_only=False)
    default_config = load_config('drs', defaults_only=True)

    missing_keys = list()

    for key in user_config.keys():
        if key not in default_config.keys():
            missing_keys.append(key)
    
    if len(missing_keys) > 0:
        raise AssertionError('The following keys are missing from the installed defaults: ' + ', '.join(missing_keys))


"""Test utility functions."""
import axiom.drs.utilities as adu

def test_is_error_recoverable():
    """Test is_error_recoverable."""
    
    rex = Exception('lock_acquire')
    urex = Exception('unrecoverable')

    recoverable_errors = [
        'lock_acquire'
    ]

    assert adu.is_error_recoverable(rex, recoverable_errors) == True
    assert adu.is_error_recoverable(urex, recoverable_errors) == False


def test_assemble_qsub_vars():
    """Test assemble_qsub_vars."""
    
    result = adu.assemble_qsub_vars(myvar1='myval1', myvar2='myval2')
    assert result == 'myvar1=myval1,myvar2=myval2'


def test_assemble_qsub_command():
    """Test assemble_qsub_command."""
    
    directives = [
        '-l walltime=%(walltime)s',
        '-q %(queue)s',
        '-N %(jobname)s',
        '-v %(qsub_vars)s'
    ]

    context = dict(
        walltime='00:30:00',
        queue='normal',
        jobname='myjob',
        qsub_vars=adu.assemble_qsub_vars(myvar1='myval1', myvar2='myval2')
    )

    result = adu.assemble_qsub_command('test.sh', directives, **context)
    expected = 'qsub -l walltime=00:30:00 -q normal -N myjob -v myvar1=myval1,myvar2=myval2 test.sh'
    
    assert result == expected


def test_filter_by_variable_name():
    """Test filter_by_variable_name."""
    
    filepaths = [
        'var1_test.nc',
        'var2_test.nc'
    ]

    # Ensure that only one file is returned
    result = adu.filter_by_variable_name(filepaths, 'var1')
    assert len(result) == 1 and result[0] == filepaths[0]


"""Automation module."""
import os
import sys
import json
import time
import datetime
from axiom.config import load_config
import axiom.utilities as au
from axiom.drs.payload import Payload
from axiom.schemas import load_schema
import axiom.drs.cli as adc


def get_instance_dir(instance):
    """Get the instance directory.

    Args:
        instance (dict): Processing instance.

    Returns:
        str: Instance directory.
    """
    return os.path.join(
        instance['work_dir'],
        instance['experiment'],
        instance['frequency']
    )


def generate_payloads(instance):
    """Generate the payloads for the processing instance.

    Args:
        instance (dict): Processing instance.
        work_dir (str): Working directory.
        experiment_base_dir (str): Base directory for the experiment.
    """
    logger = au.get_logger(__name__)

    experiment = instance['experiment']
    frequency = instance['frequency']
    logger.info(f'Generating payloads for {experiment} {frequency}')

    # Load the template payload
    payload_template = Payload.from_json(instance['payload_template'])

    # Load the schema and extract variables
    schema = load_schema(instance['schema'])
    variables = list(schema['variables'].keys())
    
    # Cut into batches of variables
    batches = au.batch_split(variables, instance['batches'])

    # Get the instance directory
    instance_dir = get_instance_dir(instance)
    payload_dst = os.path.join(
        instance_dir,
        'payloads'
    )

    os.makedirs(payload_dst, exist_ok=True)

    for year in range(instance['start_year'], instance['end_year'] + 1):
        for batch_ix, batch in enumerate(batches):
            
            # Batch identifier
            bbb = str(batch_ix).zfill(3)

            # Assemble the payload
            payload = payload_template
            payload.start_year = year
            payload.end_year = year

            # Add the variables included in this batch
            payload.variables = list(batch)
            payload.output_frequency = instance['frequency']

            # Write the payload file, with a batch identifier
            output_filename = payload.get_filename().replace('.json', f'.{bbb}.json')
            output_filepath = os.path.join(payload_dst, output_filename)
            print(output_filepath)
            payload.to_json(output_filepath)


def get_state(instance_dir, step):
    
    state_file = os.path.join(instance_dir, 'state', f'{step}.state')

    if os.path.isfile(state_file) is False:
        return False
    
    # Get the last line of the file
    with open(state_file, 'r') as f:
        lines = f.readlines()
        last_line = lines[-1]
    
    # Remove the newline character
    last_line = last_line.replace('\n', '')

    return last_line


def update_state(instance_dir, step, message):
    state_file = os.path.join(instance_dir, 'state', f'{step}.state')

    os.makedirs(os.path.dirname(state_file), exist_ok=True)

    with open(state_file, 'a') as f:
        f.write(f'{message}\n')


def check_failed(instance):
    """Check if any payloads have failed to process."""
    
    instance_dir = get_instance_dir(instance)
    
    payload_dst = os.path.join(
        instance_dir,
        'payloads'
    )

    failed_filepaths = au.auto_glob(os.path.join(payload_dst, '*.failed'), recursive=False)

    if len(failed_filepaths) > 0:
        return False


def check_consumed(instance):
    """Check that the number of payloads consumed matches the number of payloads generated.

    Args:
        instance (dict): Instance

    Returns:
        bool: True if all payloads have been consumed, False otherwise.
    """
    instance_dir = get_instance_dir(instance)
    num_payloads = len(au.auto_glob(os.path.join(instance_dir, 'payloads', '*.json')))
    num_consumed = len(au.auto_glob(os.path.join(instance_dir, 'payloads', '*.consumed')))
    return num_payloads == num_consumed


def submit_payloads(instance):

    instance_dir = get_instance_dir(instance)
    log_dir = os.path.join(instance_dir, 'logs')

    launch_kw = instance['drs_launch_kwargs']

    launch_kw['path'] = os.path.join(instance_dir, 'payloads', '*.json')
    launch_kw['log_dir'] = log_dir

    # Submit the job
    adc.drs_launch(
        **launch_kw
        # os.path.join(instance_dir, 'payloads', '*.json'),
        # instance['jobscript'],
        # log_dir,
        # dry_run=True
    )


def process(instance):


    logger = au.get_logger(__name__)
    
    # Check if payloads have been generated
    instance_dir = get_instance_dir(instance)

    # Generate payloads
    if get_state(instance_dir, 'generate_payloads') != 'success':
        generate_payloads(instance)
        update_state(instance_dir, 'generate_payloads', 'success')
    logger.info('Payloads generated')
    
    # Submit payloads
    if get_state(instance_dir, 'submit_payloads') != 'success':
        logger.info('Submitting payloads')
        submit_payloads(instance)
        update_state(instance_dir, 'submit_payloads', 'success')

    logger.info('Payloads submitted')

    # Get the number of jobs running
    logger.info('Checking job status')
    num_payloads = len(au.auto_glob(os.path.join(instance_dir, 'payloads', '*.json')))
    num_consumed = len(au.auto_glob(os.path.join(instance_dir, 'payloads', '*.consumed')))
    num_failed = len(au.auto_glob(os.path.join(instance_dir, 'payloads', '*.failed')))

    # Check if anything has failed
    if num_failed > 0:
        logger.info('Failures detected.')
        raise InstanceFailedError(f'{num_failed} payloads failed to process')
    
    # Check if all payloads have been consumed
    if num_payloads == num_consumed:
        logger.info('All payloads consumed')
        return True
    else:
        logger.info('Job is still running...')
        return False


def process_all():

    logger = au.get_logger(__name__)

    # Load the config
    logger.info('Loading automation configuration')
    automation_config = load_config('automation')
    instances = automation_config['instances']
    
    num_instances = len(instances)
    num_complete = 0
    num_failed = 0

    logger.info(f'Processing {num_instances} instances')
    for instance in instances:

        try:
            
            complete = process(instance)

            # If the instance is not complete, we do not want to move on to the next instance
            if not complete:
                logger.info('Instance incomplete')
                break
            else:
                logger.info('Instance complete')
                num_complete += 1
                continue

        # If there are any failures for this instance, move on to the next.
        except InstanceFailedError as e:
            num_failed += 1
            logger.error(e)
            continue
    
    # Resubmit only of there are still some instances to go
    if num_complete + num_failed < num_instances:
        
        resubmit_interval_seconds = automation_config['resubmit_interval_seconds']
        resubmit(resubmit_interval_seconds)
    
    else:

        logger.info('All instances complete')

        

class InstanceFailedError(Exception):
    pass    


def resubmit(interval_seconds):

    logger = au.get_logger(__name__)

    # If running under PBS then get the job submission information
    if 'PBS_JOBID' in os.environ:

        logger.info('Running under PBS, resubmitting...')

        # Get the job ID
        job_id = os.environ['PBS_JOBID']

        # Get the status of the job
        qstat_cmd = f'qstat -f -F json {job_id}'
        qstat = json.loads(au.shell(qstat_cmd, capture_output=True).stdout.decode('utf-8'))['Jobs'][job_id]
        submit_args = qstat['Submit_arguments']

        # Get the current time, add a delta for the next submission
        logger.info('Resubmitting for next interval...')
        now = datetime.datetime.now()
        dt = datetime.timedelta(seconds=interval_seconds)
        next_submit = (now + dt).strftime('%Y%m%d%H%M')

        qsub_cmd = f'qsub -a {next_submit} {submit_args}'
        qsub = au.shell(qsub_cmd, capture_output=True)

        if qsub.returncode == 0:
            logger.info('Resubmitted for next interval')
            new_jobid = qsub.stdout.decode('utf-8').strip()
            logger.info(f'New job ID: {new_jobid}')
            sys.exit()

    else:

        logger.info('Not running under PBS, resubmitting via Python.')
        logger.info(f'Sleeping for {interval_seconds}')
        time.sleep(interval_seconds)
        process_all()


if __name__ == '__main__':
    process_all()


"""Command-line methods for the QA subsystem."""
import os
import sys
import argparse
import axiom.schemas as axs
import axiom.qa as axq
from axiom.config import load_config
import axiom.drs.utilities as adu
import axiom.utilities as au
from axiom.drs.payload import Payload
from blush import parallelise, unpack_results
import numpy as np
import pandas as pd
import xarray as xr


def qa_timeseries(path, schema, start_year, end_year, output_filepath, errors=False):
    """Scan the timeseries for missing files.
    
    Args:
        path (str): Globbable path with {variable} and {year} placeholders.
        schema (str): Schema name or path.
        start_year (int): Start year.
        end_year (int): End year.
        output_filepath (str): Path to which to write the report.
        errors (bool, Optional): Report only on errors. Defaults to False.  
    """
    
    # Load the schema
    _schema = axs.load_schema(schema)
    variables = list(_schema['variables'].keys())

    print('Checking timeseries...')
    df = axq.check_timeseries_variable(path, variables, start_year, end_year)

    # Filter only for errors.
    if errors:
        df = df[df['status'] == 'ERROR']

    # Write the report
    output_filepath = os.path.abspath(output_filepath)
    df.to_csv(output_filepath, index=False)
    
    print(f'Report available at {output_filepath}')


def qc(schema, payload, start_year, end_year, report_dir=None, nstd=2.0, pct_mean=0.75, checks='nan,nstd,pct_mean', ignore_missing_inputs=False, create_payloads=False):
    """Run Quality-Control.

    Args:
        schema (str): Schema key or filepath used in processing.
        payload (str): Path to one of the payloads used in processing.
        start_year (int): First year of data.
        end_year (int): Last year of data.
        report_dir (str, optional): Path to which to write report files. Defaults to None.
        nstd (float, optional): Number of standard deviations out to consider anomalous. Defaults to 2.0.
        pct_mean (float, optional): Threshold percentage of mean file size to consider anomalous. Defaults to 0.75.
        checks (str, optional): Comma-separated checks to apply. Defaults to 'nan,nstd,pct_mean'.
        ignore_missing_inputs (bool, optional): Ignore variables that are missing from the input directory, requires that directory still exists. Defaults to False.
    """

    # Break up the checks for evaluation later
    checks = set(sorted(checks.split(',')))

    # Load the configuration, get the templates for filepaths
    config = load_config('drs')
    logger = au.get_logger(__name__)

    # Start building context
    context = config.get('metadata_defaults')
    
    drs_template = adu.get_template(config, 'drs_path')
    filename_template = adu.get_template(config, 'filename')

    # Load the schema
    _schema = axs.load_schema(schema)

    # Load the payload, update the context
    _payload = Payload.from_json(payload)
    context.update(_payload.to_dict())

    context['gcm_institute'] = '*'
    context['gcm_model'] = '*'
    context['rcm_model'] = '*'

    context['frequency_mapping'] = config.get('frequency_mapping')[_payload.output_frequency]

    expected_filepaths = list()
    variables = list()
    years = list()

    # Create a list of variables to check
    variables2check = _schema['variables'].keys()
    variables2ignore = list()

    # Get a list of variables that are in the input directory to exclude.
    input_variables = list()

    logger.info('Assembling a list of variables to check.')

    if ignore_missing_inputs:

        logger.info('User has requested missing inputs be ignored')

        input_filepaths = au.auto_glob(_payload.input_files)
        if len(input_filepaths) == 0:
            logger.error('Inputs no longer exist! Unable to proceed with QC using --ignore_missing_inputs')
            raise FileNotFoundError('Unable to collect a list of variables from payload inputs, does the directory still exist?')
        
        for input_filepath in input_filepaths:
            
            input_variable = os.path.basename(input_filepath).split('_')[0]
            
            # Skip variables that have previously been ignored / added.
            if input_variable in variables2ignore or input_variable in input_variables:
                continue

            # On the first instance, check if it is not expected due to frequency
            if input_variable not in input_variables:

                logger.info(f'Checking if we should test for {input_variable}')
                
                ds = xr.open_dataset(input_filepath, chunks=dict(time=1))
                freq = adu.detect_input_frequency(ds)
            
                if freq != _payload.output_frequency and config.allow_subdaily_resampling == False:
                    logger.info(f'{input_variable} is on a different frequency and allow_subdaily_resampling is disabled. Ignoring')
                    variables2ignore.append(input_variable)
                    continue

            logger.info(f'Adding {input_variable} to the list to be checked.')                    
            input_variables.append(input_variable)

        # Perform an intersection to get the true list of variables to check.
        variables2check = list(set(variables2check) & set(input_variables))

    # Loop through all of the variables
    for variable in variables2check:
        
        for year in range(start_year, end_year+1):

            context['variable'] = variable
            context['start_date'], context['end_date'] = adu.get_start_and_end_dates(year, _payload.output_frequency)
            drs_path = drs_template % context
            filename = filename_template % context
            expected_filepath = os.path.join(drs_path, filename)
            expected_filepaths.append(expected_filepath)
            variables.append(variable)
            years.append(year)
    
    # Check all of the files in parallel
    logger.info(f'Checking timeseries...')
    results = parallelise(_check_file, num_threads=8, filepath=expected_filepaths, year=years, variable=variables)
    results = unpack_results(results)

    # Assemble a dataframe for aggregate statistics
    df = pd.DataFrame(results)

    df_nstd = None
    df_nan = None
    df_pct_mean = None

    # Check the results
    for variable in df.variable.unique().tolist():

        # Extract just this variable
        var_df = df[df.variable == variable]

        # Do nan check
        if 'nan' in checks:
            _df_nan = _filter_nan(var_df)
            df_nan = _df_nan if df_nan is None else pd.concat([df_nan, _df_nan])

        # Do nstd check
        if 'nstd' in checks:
            _df_nstd = _filter_nstd(var_df, nstd)
            df_nstd = _df_nstd if df_nstd is None else pd.concat([df_nstd, _df_nstd])
        
        # Do pct_mean check
        if 'pct_mean' in checks:
            _df_pct_mean = _filter_pct_mean(var_df, threshold=pct_mean)
            df_pct_mean = _df_pct_mean if df_pct_mean is None else pd.concat([df_pct_mean, _df_pct_mean])


    # Check if there are any errors to report
    num_nan = len(df_nan.index) if df_nan is not None else 0
    num_nstd = len(df_nstd.index) if df_nstd is not None else 0
    num_pct_mean = len(df_pct_mean.index) if df_pct_mean is not None else 0

    # Take the sum of the entries for the error check
    is_error = num_nan + num_nstd + num_pct_mean > 0

    # Exit with status
    if is_error:
        logger.error('Timeseries has failed QC')
        logger.error(f'{num_nan} Files missing/no filesize = {num_nan}')
        logger.error(f'{num_nstd} files are more than {nstd} away from the mean file size.')
        logger.error(f'{num_pct_mean} files are less than {pct_mean} from the mean file size.')
    
    # Write the reports at the given directory
    if report_dir and is_error:
        os.makedirs(report_dir, exist_ok=True)

        if 'nstd' in checks:
            df_nstd.to_csv(os.path.join(report_dir, 'errors_nstd.csv'))
        
        if 'nan' in checks:
            df_nan.to_csv(os.path.join(report_dir, 'errors_nan.csv'))
        
        if 'pct_mean' in checks:
            df_pct_mean.to_csv(os.path.join(report_dir, 'errors_pct_mean.csv'))
        
        logger.error(f'Reports written to {report_dir}')

    # Create payloads to rerun for the errors
    if create_payloads and is_error:

        for year, group_df in df_nan.groupby('year'):

            rerun_payload = _payload
            rerun_payload.start_year = year
            rerun_payload.end_year = year
            rerun_payload.variables = group_df.variable.unique().tolist()

            # print(rerun_payload.to_dict())
            payload_filepath = os.path.join(
                report_dir,
                rerun_payload.get_filename()
            )

            rerun_payload.to_json(payload_filepath)

    sys.exit(int(is_error))


def _filter_nstd(df, nstd):
    """Check the filesize for anomalies that exceed nstd from the mean.

    Args:
        df (pandas.Dataframe): Dataframe with a 'size' column.
        nstd (float): Number of standard deviations to consider.
    
    Returns:
        pandas.Dataframe: Dataframe filtered by those that exceed nstd, with additional column 'nstd_from_mean_size'.
    """
    _mean = df['size'].mean()
    _std = df['size'].std()

    # Get the distance away from the mean, check the threshold
    df['nstd_from_mean'] = (df['size'] - _mean) / _std
    df = df[(np.abs(df['nstd_from_mean_size']) > nstd)]

    return df


def _filter_pct_mean(df, threshold):
    """Filter the files for those whose size is less that threshold of the mean.

    Args:
        df (pandas.Dataframe): Dataframe with a 'size' column.
        threshold (float): Fraction to below which will constitute an anomaly.
    
    Returns:
        pandas.Dataframe: Dataframe filtered by anomalies less than the fractional threshold supplied, with additional column 'pct_mean_size'.
    """
    _mean = df['size'].mean()
    df['pct_mean_size'] = df['size'] / _mean
    df = df[df.pct_mean_size <= threshold]
    return df


def _filter_nan(df):
    """Filter the files for those which have NaN for their filesize.

    Args:
        df (pandas.Dataframe): Dataframe with a 'size' column.
        threshold (float): Fraction to below which will constitute an anomaly.
    
    Returns:
        pandas.Dataframe: Dataframe filtered by anomalies less than the fractional threshold supplied, with additional column 'pct_mean_size'.
    """
    return df[df['size'].isnull()]


def _check_file(filepath, year, variable):

    found_files = au.auto_glob(filepath)

    result = dict(year=year, variable=variable)

    # Check if the variable is actually fixed?
    fixed_path = os.path.join(
        os.path.dirname(os.path.dirname(os.path.dirname(filepath))),
        'fx',
        variable,
        f'{variable}_*.nc'
    )

    found_fixed_files = au.auto_glob(fixed_path)

    if not found_files and not found_fixed_files:
        result.update(dict(filepath=filepath, size=np.nan))
        return result
    
    if found_files:
        _filepath = found_files[0]
    elif found_fixed_files:
        _filepath = found_fixed_files[0]

    if not _filepath:
        result.update(dict(filepath=_filepath, size=np.nan))
        return result
    
    size = os.path.getsize(_filepath)
    result.update(dict(filepath=_filepath, size=size))
    return result


def get_parser(config=None, parent=None):
    """Parse arguments for the QA module.

    Args:
        config (dict, optional): Configuration dictionary. Defaults to None.
        parent (obj, optional): Parent parser object. Defaults to None.
    
    Returns:
        argparse.Namespace: Arguments object.
    """
    if parent is None:
        parser = argparse.ArgumentParser()
    else:
        parser = parent.add_parser('drs_qc')
    
    parser.add_argument('schema', type=str, help='Schema name or path that was used to process data.')
    parser.add_argument('payload', type=str, help='Example payload that was used to process data.')
    parser.add_argument('start_year', type=int, help='Start year.')
    parser.add_argument('end_year', type=int, help='End year.')
    parser.add_argument('--report_dir', type=str, help='Optional directory in which to place error reports.', default=None)
    parser.add_argument('--nstd', type=float, help='Number of standard deviations out to consider an error. (Default = 2.0)', default=2.0)
    parser.add_argument('--pct_mean', type=float, help='Percentage of mean file size out to consider an error (fraction). (Default = 0.75)', default=0.75)
    parser.add_argument('--checks', type=str, help='Checks to run. Defaults to "nan,nstd,pct_mean"', default='nan,std,pct_mean')
    parser.add_argument('--ignore_missing_inputs', help='Ignore variables that are not found in the input directory (requires that directory still exist!)', action='store_true', default=False)
    parser.add_argument('--create_payloads', help='Create payloads to rerun for the different errors.', action='store_true', default=False)
    parser.set_defaults(func=qc)
    
    return parser



"""Command-line methods for the DRS subsystem."""
from ast import arg
import os
import sys
import json
import argparse
import axiom.utilities as au
from axiom.config import load_config
import axiom.drs.payload as adp
import axiom.drs.utilities as adu
from tqdm import tqdm
from pathlib import Path
import shutil
import datetime


def split_args(values):
    """Split an argument that is comma-separated.

    Args:
        values (str): Values
    
    Returns:
        list : List of split arguments.
    """
    return values.split(',')


def get_parser(config=None, parent=None):
    """Parse arguments for command line utiltities.

    Args:
        config (dict) : Configuration dictionary.
        parent (obj) : Parent parser object (for integration into the main axiom CLI)

    Returns:
        argparse.Namespace : Arguments object.
    """
    
    # Load the model, project and domain
    VALID_MODELS = load_config('models').keys()
    VALID_PROJECTS = load_config('projects').keys()
    VALID_DOMAINS = load_config('domains').keys()

    # Build a parser
    if parent is None:
        parser = argparse.ArgumentParser()
    # ...or add one to the top-level CLI
    else:
        parser = parent.add_parser('drs')

    parser.description = "DRS utility"

    # Paths
    parser.add_argument('--input_files', required=True, type=str, help='Input filepaths', nargs=argparse.ONE_OR_MORE)
    parser.add_argument('--output_directory', required=True, type=str, help='Output base directory (DRS structure built from here)')
    parser.add_argument('-o', '--overwrite', default=False, help='Overwrite existing output', action='store_true')

    # Temporal
    parser.add_argument('-s', '--start_year', required=True, type=int, help='Start year')
    parser.add_argument('-e', '--end_year', required=True, type=int, help='End year')

    # Resolution and output frequency
    parser.add_argument('-r', '--input_resolution', type=float, help='Input resolution in km, leave blank to auto-detect from path.')
    
    parser.add_argument(
        '-f', '--output_frequency', required=True, type=str, metavar='output_frequency',
        help='Output frequency, Examples include "12min", "1M" (1 month) etc. see https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases.'
    )

    # Metadata
    parser.add_argument('-p', '--project', required=True, type=str, choices=VALID_PROJECTS)
    parser.add_argument('-m', '--model', required=True, type=str, choices=VALID_MODELS)

    # Domains, we can process multiple at once
    parser.add_argument(
        '-d', '--domain',
        required=True, type=str,
        choices=VALID_DOMAINS,
        help='Domain to process'
    )

    # Override the variables defined in the drs.json file
    parser.add_argument(
        '-v', '--variable',
        type=str,
        required=True,
        help='Variable to process.'
    )

    return parser


def get_parser_consume(config=None, parent=None):
    """A parser that can consume JSON payload files.

    Args:
        config ([type], optional): [description]. Defaults to None.
        parent ([type], optional): [description]. Defaults to None.
    """
    if parent is None:
        parser = argparse.ArgumentParser()
    else:
        parser = parent.add_parser('drs_consume')

    # Input filepaths
    parser.add_argument('input_filepaths', type=str, help='Input json filepaths.', nargs=argparse.ONE_OR_MORE)
    return parser


def drs_launch(path, jobscript, log_dir, batches=None, dry_run=True, interactive=False, unlock=False, **launch_context):
    """Method to launch a series of qsubs for DRS processing.

    Args:
        path (str): Globbable path of payload files.
        jobscript (str): Path to the job script.
        log_dir (str): Path to which to save the log files.
        batches (int): Number of batches to split variables into (for parallel processing).
        dry_run (bool): Print out the commands rather than executing.
        interactive (bool): Dump the interactive flag into the qsub command when dumping.
        unlock (bool): Unlock locked payloads prior to submission (for rerunning walltime overruns)
        **launch_context: Additional arguments that will be interpolated as launch context.
    """

    # List the payloads in the input_directory
    payloads = au.auto_glob(path)

    if not dry_run:
        os.makedirs(log_dir, exist_ok=True)

    for payload in payloads:

        # Load the payload to get the project, where we can get the variables and work out what the batch_size will be
        if batches:
            payload_obj = json.load(open(payload, 'r'))
            project = payload_obj['project']
            project_config = load_config('projects')[project]
            variables = project_config['variables_2d'] + list(project_config['variables_3d'].keys())
            variables_batches = au.batch_split(variables, n_batches=batches)
            _batches = range(1, len(variables_batches) + 1)
            
        
        # Single batch, which makes the next section simpler
        else:
            _batches = [1]

        # Unlock the file if requested
        if au.is_locked(payload) and unlock == True:
            print(f'Unlocking {payload} for resubmission')
            au.unlock(payload)
        
        # Skip if not
        elif au.is_locked(payload):
            print(f'{payload} is locked.')
            continue

        # Convert the path to the jobscript to an absolute path for reproducibility
        jobscript = os.path.abspath(jobscript)
        payload = os.path.abspath(payload)
        log_dir = os.path.abspath(log_dir)

        for batch_id in _batches:

            job_name = os.path.basename(payload)
            batch_str = str(batch_id).zfill(3)
            job_name = f'{job_name}_{batch_str}'

            # Assemble the command from configuration
            config = load_config('drs')
            directives = config['launch']['directives']

            # Add interactive flag when dry running
            if dry_run and interactive:
                directives.append('-I')

            # Override walltime
            if 'walltime' in launch_context.keys() and launch_context['walltime'] is not None:
                walltime = launch_context['walltime']
                directives.append(f'-l walltime={walltime}')

            qsub_vars = dict(
                AXIOM_PAYLOAD=payload,
                AXIOM_LOG_DIR=log_dir,
                AXIOM_BATCH=batch_id
            )

            # Assemble the launch context for this job
            _launch_context = dict(
                qsub_vars=adu.assemble_qsub_vars(**qsub_vars),
                job_name=job_name,
                log_dir=log_dir,
                batch_str=batch_str
            )

            # Add this to the user-supplied launch context
            launch_context.update(_launch_context)

            # Assemble the qsub command
            cmd = adu.assemble_qsub_command(
                jobscript=jobscript,
                directives=directives,
                **launch_context
            )

            # Dry run, just echo the outputs
            if dry_run:
                print(cmd)
        
            # Real run, submit the jobs.
            else:
                qsub = au.shell(cmd)
                if qsub.returncode == 0:
                    print(qsub.stdout.decode('utf-8'))


def get_parser_launch(parent=None):
    """A parser that can launch DRS processing.

    Args:
        parent ([type], optional): [description]. Defaults to None.
    """
    if parent is None:
        parser = argparse.ArgumentParser()
    else:
        parser = parent.add_parser('drs_launch')

    parser.description = 'Submit drs_consume tasks via qsub.'

    # Input filepaths
    parser.add_argument('path', type=str, help='Globbable path to payload files (use quotes)')
    parser.add_argument('jobscript', type=str, help='Path to the jobscript for submission.')
    parser.add_argument('log_dir', type=str, help='Directory to which to write logs.')
    parser.add_argument('-d', '--dry_run', action='store_true', default=False, help='Print commands without executing.')
    parser.add_argument('-i', '--interactive', action='store_true', default=False, help='Dump the interactive flag into the qsub command when dry-running.')
    parser.add_argument('--walltime', type=str, help='Override walltime in job script.')
    parser.add_argument('--unlock', help='Unlock locked payloads prior to submission', action='store_true', default=False)
    parser.set_defaults(func=drs_launch)

    return parser


def generate_payloads(payload_dst, input_files, output_dir, start_year, end_year, project, model, domain, variables=None, schema=None, output_frequencies='1H,6H,1D,1M', num_batches=1, extra=None):
    
    # Unpack the extra arguments
    _extra = dict()
    for kv in extra:
        k, v = kv.split(',')
        _extra[k] = v

    payloads = adp.generate_payloads(
        input_files=input_files,
        output_directory=output_dir,
        start_year=start_year, end_year=end_year,
        project=project, model=model, domain=domain,
        variables=variables, schema=schema,
        output_frequencies=output_frequencies,
        num_batches=num_batches,
        **_extra
    )

    if len(payloads) == 0:
        raise Exception('No payloads generated!!!')

    # Create the output directory
    os.makedirs(payload_dst, exist_ok=True)

    # Generate the payloads at the destination
    print('Writing payloads...')
    for payload in tqdm(payloads):
        
        filepath = os.path.join(
            payload_dst,
            payload.get_filename()
        )

        payload.to_json(filepath)
    
    print(f'Payloads available at {payload_dst}')

        


def get_parser_generate_payloads(parent=None):
    """A parser to generate payloads.

    Args:
        parent (object, optional): Parent parser object. Defaults to None.
    """
    parser = argparse.ArgumentParser() if parent is None else parent.add_parser('drs_gen_payloads')

    parser.description = 'Generate a series of payload files.'
    parser.add_argument('payload_dst', type=str, help='Where to write payload files.')
    parser.add_argument('input_files', type=str, help='Globbable path to input files, use quotes.')
    parser.add_argument('output_dir', type=str, help='Output directory (DRS written from here).')
    parser.add_argument('start_year', type=int, help='Start year.')
    parser.add_argument('end_year', type=int, help='End year.')
    parser.add_argument('project', type=str, help='Project key from projects.json.')
    parser.add_argument('model', type=str, help='Model key from models.json.')
    parser.add_argument('domain', type=str, help='Domain key from domains.json.')

    parser.add_argument('--variables', type=split_args, help='Comma-separated list of variables to process.')
    parser.add_argument('--schema', type=str, help='Schema to read variables from in lieu of variables.')
    parser.add_argument('--output_frequencies', type=split_args, help='Comma-separated list of output frequencies. Defaults to "1H,6H,1D,1M"', default='1H,6H,1D,1M')

    parser.add_argument('-e', '--extra', type=str, nargs=argparse.ZERO_OR_MORE, help='Extra metadata to add, "key,value".')
    parser.set_defaults(func=generate_payloads)

    return parser


def rerun_failures(input_dir):
    """Method to rerun payloads based on the .failed output files.
    
    Args:
        input_dir (str) : Path to the input directory containing both the failed files and the original payloads.
    """

    # Find all of the failed files in the provided directory
    failed_filepaths = au.auto_glob(f'{input_dir}/*failed')

    for failed_filepath in failed_filepaths:

        # Open the failed filepath, read out the variables
        raw = open(failed_filepath).read().splitlines()
        failed_variables = [_raw.split(',')[0] for _raw in raw]

        # Filter out duplicates
        failed_variables = list(set(failed_variables))

        # Open the payload
        payload_filename = os.path.basename(failed_filepath).split('_')[0]
        payload_filepath = os.path.join(
            input_dir,
            payload_filename
        )

        # Open them
        payload = adp.Payload.from_json(payload_filepath)

        # Replace the variables listed with that of those included in the failed file
        payload.variables = failed_variables
        
        # Save them to a rerun directory
        output_dir = os.path.join(input_dir, 'rerun')
        os.makedirs(output_dir, exist_ok=True)

        output_filepath = os.path.join(
            output_dir,
            payload_filename
        )

        print(output_filepath)
        payload.to_json(output_filepath)


def get_parser_rerun_failures(parent=None):
    """Get a parser for rerunning payloads.

    Args:
        parent (object, optional): Parent parser. Defaults to None.
    """
    parser = argparse.ArgumentParser() if parent is None else parent.add_parser('drs_rerun_failures')
    parser.description = 'Generate rerun payloads from the .failed files in the input directory'
    parser.add_argument('input_dir', type=str, help='Path to .failed files and their payloads.')
    parser.set_defaults(func=rerun_failures)
    return parser


def get_parser_generate_user_config(parent=None):
    """Get a parser for generating a set of user config files from the installation directory.

    Args:
        parent (object, optional): Parent parser. Defaults to None.
    """
    parser = argparse.ArgumentParser() if parent is None else parent.add_parser('drs_gen_user_config')
    parser.description = 'Copy installation configuration to the user space (backing up anything already there).'
    parser.set_defaults(func=adu.generate_user_config)
    return parser


"""Configuration."""
import os
import json
import pkgutil
import pathlib
import axiom.utilities as au


class Config(dict):
    

    def __getattr__(self, key, default=None):
        """Allows attribute access for top-level keys.

        Args:
            key (str): Key.

        Returns:
            Any: Value.
        """
        return self.__getitem__(key)


    def __setattr__(self, key, value):
        """Allows attribute access for top-level keys.

        Args:
            key (str): Key.
            value (Any): Value to assign.
        """
        super().__setitem__(key, value)

    def __getitem__(self, key):
        """Boolean fallback method, returning False if key does not exist.

        Use for adding new conditional functionality.

        Args:
            key (hashable): Key

        Returns:
            object : Value in config, or False
        """
        
        try:
            value = super().__getitem__(key)
            
        except KeyError:
            value = False

        return value

    def get(self, key, default=None):
        
        if key in self.keys():
            return self[key]

        if default:
            return default

        raise KeyError(f'Config key {key} does not exist.')
        

    def load(self, config_name, defaults_only=False):
        """Load the configuration in a cascading fashion, defaults first, then overlay with user.

        Args:
            config_name (str): Configuration name, without file extension.
            defaults_only (bool, optional): Load only the defaults. Defaults to False.
        """

        default_filepath = os.path.join(au.get_installed_data_root(), f'{config_name}.json')
        user_filepath = os.path.join(au.get_user_data_root(), f'{config_name}.json')
        
        # Load any installed defaults, if they exists
        if os.path.isfile(default_filepath):
            defaults = json.load(open(default_filepath, 'r'))
        else:
            defaults = dict()

        if defaults_only:
            self.update(defaults)
            return

        # Load the user configuration over the top
        if os.path.isfile(user_filepath):
            user = json.load(open(user_filepath, 'r'))
            defaults.update(user)

        self.update(defaults)


def load_config(config_name, defaults_only=False):
    """Shorthand to load a configuration object.

    Args:
        config_name (str): Name of the config file.
        defaults_only (bool, optional): Load only the defaults. Defaults to False.

    Returns:
        axiom.Config: Configuration object.
    """
    config = Config()
    config.load(config_name, defaults_only=defaults_only)
    return config


"""Tests for the DRS subsystem."""
import re
import axiom.drs.utilities as adu



def test_parse_domain():
    """Test parsing a domain."""
    directive = "Antarctica,0.5,-90.0,-50.0,-180.0,180.0"

    expected = dict(
        name='Antarctica',
        dx=0.5,
        lat_min=-90.0, lat_max=-50.0,
        lon_min=-180.0, lon_max=180.0
    )

    result = adu.parse_domain(directive)
    assert result == expected


def test_get_uninterpolated_placeholders():
    """Test uninterpolated value check."""
    
    expect_something = 'this is %(a)s test of %(b)s'
    result = adu.get_uninterpolated_placeholders(expect_something)
    assert result == ['a', 'b']

    expect_nothing = 'this is a test of b'
    result = adu.get_uninterpolated_placeholders(expect_nothing)
    assert len(result) == 0


"""Validator class."""
import axiom.utilities as au
from cerberus import Validator as CerberusValidator
from datetime import datetime


class Validator:

    """Validation object:

    Args:
        schema (str, dict) : Path to schema or schema dictionary.
    """

    def __init__(self, schema):

        self.is_valid = None
        self.errors = dict(_global=list(), variables=dict())
        self.schema_filepath = None
        self.date_validated = None

        # Ensure schema is valid typew
        assert isinstance(schema, str) or isinstance(schema, dict)

        # Allow a path to be specified
        if isinstance(schema, str):
            self.schema_filepath = schema
            schema = au.load_schema_json(schema)

        self.schema = schema


    def validate(self, metadata, allow_unknown=True):
        """Validate metadata against a schema.

        Args:
            metadata ([type]): [description]
            allow_unknown (bool, optional): [description]. Defaults to True.

        Returns:
            bool : True if valid, False otherwise. Errors are stored in errors attribute.
        """

        # Reset the errors and valid status
        self.is_valid = True
        self.errors = dict(_global=list(), variables=dict())
        self.date_validated = datetime.utcnow()

        # Validate the global attributes
        v = CerberusValidator(schema=self.schema['_global'], allow_unknown=allow_unknown, require_all=True)
        if not v.validate(metadata['_global']):
            self.errors['_global'] = v._errors
            self.is_valid = False

        # Set up a default variable schema, to enfore a minimum standard
        if '_default' in self.schema['variables'].keys():
            default_schema = self.schema['variables']['_default']
        else:
            default_schema = dict()

        # Validate each variable one by one
        for k, attrs in metadata['variables'].items():

            # Apply the default schema, overwrite with the variable-specific one
            _schema = default_schema

            # Skip unknown variables if permitted
            if k not in self.schema['variables'].keys() and allow_unknown == False:
                continue

            # Apply this variable's schema to the default if it is defined
            if k in self.schema['variables'].keys():
                _schema.update(self.schema['variables'][k])

            v = CerberusValidator(
                schema=_schema,
                allow_unknown=allow_unknown,
                require_all=True
            )

            if not v.validate(attrs):
                self.errors['variables'][k] = v._errors
                self.is_valid = False

        return self.is_valid



"""Entrypoint for the schemas subpackage."""
import os
import glob
import json
from importlib.metadata import version
from axiom.utilities import get_installed_data_root, get_user_data_root


# Order dependent, user to override system
SCHEMA_DIRS = [
    os.path.join(get_user_data_root(), 'schemas'),
    os.path.join(get_installed_data_root(), 'schemas')
]


def list_schema_filepaths():
    """List the schema filepaths.

    Returns:
        list: List of filepaths
    """
    schema_filepaths = list()
    # Get the installed schemas first, then the user schemas
    for schema_dir in SCHEMA_DIRS:
        if os.path.isdir(schema_dir):
            schema_filepaths += sorted(glob.glob(os.path.join(schema_dir, '*.json')))
    
    return schema_filepaths
    

def load_schemas():
    """Load all of the schemas into a dictionary object.

    Returns:
        dict : Dictionary of schemas
    """
    schemas = dict()
    for schema_filepath in list_schema_filepaths():
        schema = load_schema_file(schema_filepath)

        # Note, user schemas will override system schemas
        schemas[schema['name']] = schema
    
    return schemas


def load_schema(key_or_filepath):
    """Load the schema.

    Args:
        key_or_filepath (str): Schema key or filepath.
    
    Returns:
        dict : Schema dictionary.
    """

    # Load direct from the dictionary
    schemas = load_schemas()
    if key_or_filepath in schemas.keys():
        return schemas[key_or_filepath]

    # Attempt to load the file directly
    return load_schema_file(key_or_filepath)


def load_schema_file(filepath):
    """Actually load a schema filepath.

    Args:
        filepath (str): Path to the schema file.
    
    Returns:
        dict : Schema.
    """
    return json.loads(open(filepath, 'r').read())
    

def list_schemas():
    """List the available schemas (the json filepath).

    Returns:
        list : List of schema files available for loading.
    """
    # List everything in the schemas folder
    schemas = list_schema_filepaths()

    # Just return the basenames
    return [os.path.basename(s) for s in schemas]


"""Pre and post-processing functions for CCAM."""
import numpy as np
import datetime
import axiom.drs.utilities as adu
import axiom.utilities as au


def _detect_version(ds):
    """The CCAM version can be detected from the history metadata.

    Args:
        ds (xarray.Dataset): Dataset
    
    Returns:
        str : Version.
    """
    history = ds.attrs['history']
    yymm = datetime.datetime.strptime(history.split()[2], '%Y-%m-%d').strftime('%y%m')
    return yymm


def _set_version_metadata(ds, version):
    """Set the version metadata on the DataSet.

    Args:
        ds (xarray.DataSet): Dataset.
        version (str): Version string.

    Returns:
        xarray.Dataset : Dataset with version metadata on it.
    """
    ds.attrs['rcm_model'] = f'CCAM-{version}'
    ds.attrs['rcm_model_cordex'] = f'CCAM-{version}'
    ds.attrs['rcm_model_version'] = version
    ds.attrs['rcm_version'] = version
    ds.attrs['rcm_version_cordex'] = version
    return ds


def preprocess_ccam(ds, **kwargs):
    """Preprocess the data upon loading for CORDEX requirments.

    Args:
        ds (xarray.Dataset): Dataset.
        variable (str): Variable to extract along with bnds. Must be used as part of a lambda in open_mfdataset

    Returns:
        xarray.Dataset: Dataset with preprocessing applied.
    """

    variable = kwargs['variable']

    # Rename metadata keys if needed
    if 'rlat0' in ds.attrs.keys():
        ds.attrs['rlon'] = ds.attrs.pop('rlong0')
        ds.attrs['rlat'] = ds.attrs.pop('rlat0')

    # Automatically detect version from inputs
    if 'model_id' not in kwargs['kwargs'].keys():
        version = _detect_version(ds)
    else:
        version = kwargs['kwargs']['model_id'].split('-')[-1]
        
    ds = _set_version_metadata(ds, version)
        
    # Extract the lat/lon bounds as well.
    if variable:
        ds = ds[[variable, 'lat_bnds', 'lon_bnds']]

    return ds

def center_times(ds, output_frequency):
    """Centers the times in the dataset.

    Args:
        ds (xarray.Dataset): Data.
    
    Returns:
        xarray.Dataset : Data with times centered.
    """

    # non-monthly data is simple, just halve the delta
    if output_frequency != '1M':
        dt = ds.time.data[1] - ds.time.data[0]
        ds['time'] = ds.time + (dt / 2)
        return ds

    # Otherwise, we need to apply more logic to the problem.
    dt = ds.time.data[1:] - ds.time.data[0:-1]
    new_times = ds.time.data[:] + (dt / 2)
    new_times = np.append(new_times, new_times[6]) # july
    ds['time'] = new_times


def postprocess_ccam(ds, **kwargs):
    """For CORDEX processing, there is some minor postprocessing that happens.

    Args:
        ds (xarray.Dataset): Data.

    Returns:
        xarray.Dataset: Data with postprocessing applied.
    """

    logger = au.get_logger(__name__)

    # Strip out the extra metadata keys (Marcus 20220802)
    remove_keys = 'ensemble,rcm_institute,rcm_model_cordex,rcm_model,rcm_version_cordex'.split(',')
    for rk in remove_keys:
        if rk in ds.attrs.keys():
            logger.debug(f'Removing metadata key {rk}')
            ds.attrs.pop(rk)

    # Strip out the extra dimensions from bnds (reduces filesize considerably)
    if 'lat_bnds' in ds.data_vars.keys():
        
        # Drop surplus coordinates
        ds['lat_bnds'] = au.isolate_coordinate(ds.lat_bnds, 'lat', drop=True)
        ds['lon_bnds'] = au.isolate_coordinate(ds.lon_bnds, 'lon', drop=True)

    # Center the times for non-instantaneous data.
    _is_instantaneous = is_instantaneous(ds, kwargs['variable'])
    _resampling_applied = kwargs['resampling_applied']
    
    logger.debug(f'is_instantaneous = {_is_instantaneous}')
    logger.debug(f'resampling_applied = {_resampling_applied}')
    if _resampling_applied == True:
        logger.debug('TIME CENTERING TRIGGERED')
        ds = center_times(ds, output_frequency=['output_frequency'])

    return ds


def is_instantaneous(ds, variable):
    """Checks for the presence of CCAM-specific flags indicating that a variable is instantaneous.

    Args:
        ds (xarray.Dataset): Data.
        variable (str): Variable currently being processed.
    """

    da = ds[variable]

    # if cell_methods is missing
    if 'cell_methods' not in da.attrs.keys():
        return True
    
    # time: point is present
    if da.attrs['cell_methods'] == 'time: point':
        return True
    
    return False
    


"""Code to perform DRS functions."""
from axiom.utilities import load_package_data
import argparse
import xarray as xr
import os
import uuid
from datetime import datetime
import sys
from pprint import pprint
import re


def get_domains(resolution, frequency, variable_fixed, no_frequencies):
    """Get the domains for the arguments provided.

    Args:
        resolution (int): Resolution
        frequency (str): Frequency
        variable_fixed (bool): Is the variable fixed?
        no_frequencies (bool): True if no frequencies were provided.

    Raises:
        Exception: When the arguments provided don't yield any domain information.

    Returns:
        dict: Dictionary of domain information.
    """

    is_cordex = frequency == 'cordex'

    domains, frequencies, resolution_dir, degrees = None, None, None, None

    # Quickly bail if necessary
    assert resolution in [50, 5, 10, 12]

    if resolution == 50:

        degrees = 0.5
        resolution_dir = f'{resolution}km'

        if not is_cordex:
            domains = ['AUS-50']

        else:
            
            domains = ['AUS-44i']
            frequencies = ['day', 'month']
            
            if variable_fixed:
                frequencies = ['day']
        
        if no_frequencies:
            frequencies = ['3hr', 'day', 'month']

    elif resolution == 5:

        domains = ['VIC-5']
        degrees = 0.05
        resolution_dir = f'{resolution}km'

    elif resolution == 10:

        domains = ['SEA-10', 'TAS-10']
        resolution_dir = '5km',
        degrees = 0.1

    elif resolution == 12:

        if not is_cordex:
            raise Exception('Domain not known.')

        domains = ['AUS-44i']
        frequencies = ['day', 'month']

        if variable_fixed:
            frequencies = ['day']
    
    # Return everything
    return dict(
        domains=domains,
        frequencies=frequencies,
        resolution_dir=resolution_dir,
        degrees=degrees
    )


def get_rundir():
    """Get the run directory from either PBS or pwd.

    Returns:
        str : Run directory.
    """
    if 'PBS_O_WORKDIR' in os.environ.keys():
        return os.getenv('PBS_O_WORKDIR')

    return os.getcwd()


def get_decades(start_year, end_year):
    """Get a list of decades to process (usually only 1)

    Args:
        start_year (int) : Start year.
        end_year (int) : End year.

    Returns:
        list : List of decades to process.
    """
    
    # Convert to decades
    start_decade = start_year // 10
    end_decade = end_year // 10

    # Degenerate case, just do a single decade
    if start_decade == end_decade:
        return [start_decade]

    return range(start_decade, end_decade + 1)


def get_template(config, key):
    """Get an interpolation template out of the config.

    Args:
        config (dict): Dictionary.
        key (str): Template key.

    Returns:
        str: Template.
    """
    # Get the template out of the configuration
    template = config['templates'][key]

    # Convert list to string (helps with really long templates)
    if isinstance(template, list):
        
        # The delimiter is the first item in the list
        template = template[0].join(template[1:])

    return template


def get_lengths(decade, s_year, e_year, context=None):
    """Get the number of times expected given a decade.

    Args:
        decade (int) : Decade.
    """

    # Overrides for specific conditions
    if decade == 200:

        if s_year == 2006:
            e_year = 2009
        else:
            s_year = 2000

            if context['gcm_model'] == 'ERAINT':
                e_year = (decade * 10) + 9
            else:
                e_year = 2005
    
    elif context['gcm_model'] == 'ERAINT' and decade == 201:
        e_year = (decade * 10) + 4
    else:
        s_year = decade * 10
        e_year = (decade * 10) + 9

    
    dt =  datetime(e_year, 12, 31) - datetime(s_year, 1, 1)

    # Work out the expected lengths at different resolutions
    num_seconds = dt.total_seconds()
    num_minutes = num_seconds / 60
    num_1hr = num_minutes / 60
    num_12min = num_1hr * 5
    num_days = num_1hr / 24
    num_3hrs = num_days * 8
    num_months = ((e_year - s_year) + 1) * 12

    # Return everything
    return {
        'num_month': num_months,
        'num_day': num_days,
        'num_3hr': num_3hrs,
        'num_1hr': num_1hr,
        'num_12min': num_12min,
        'num_minutes': num_minutes,
        'num_seconds': num_seconds,
        's_year': s_year,
        'e_year': e_year,
    }


def get_lengths_new(decade, **context):
    """Get lengths.

    Args:
        **context : Context dictionary, because there are a lot of variables.
    
    Returns:
        dict : Dictionary of lengths
    """
    s_year, e_year = context['s_year'], context['e_year']
    even = decade % 2 == 0
    single_year = s_year == e_year

    lengths = dict()

    if single_year:
        
        lengths['num_month'] = 12
        lengths['num_day'] = 365

        # Check leap year
        if s_year % 4 == 0:
            lengths['num_day'] = 366

    else:

        if even:

            # There is an extra leap year in these decades
            lengths['num_day'] = 3653
            lengths['num_1hr'] = 87672

        else:

            lengths['num_day'] = 3652
            lengths['num_1hr'] = 87648
        
        if decade == 200:

            if s_year == 2006:

                e_year = 2009
                lengths['num_month'] = 48
                lengths['num_day'] = 1461
                lengths['num_1hr'] = 35064
            
            else:

                s_year = 2000

                if context['gcm_model'] == 'ERAINT':
                    
                    e_year = (decade * 10) + 9
                
                else:

                    e_year = 2005
                    lengths['num_month'] = 72
                    lengths['num_day'] = 2192
                    lengths['num_1hr'] = 52608
        
        elif context['gcm_model'] == 'ERAINT' and decade == 201:

            e_year = (decade * 10) + 4
            lengths['num_month'] = 60
            lengths['num_day'] = 1826
            lengths['num_1hr'] = 43824

        else:

            s_year = decade * 10
            e_year = (decade * 10) + 9
        
        lengths['num_12min'] = lengths['num_1hr'] * 5
        lengths['num_3hr'] = lengths['num_day'] * 8
    
    # Copy in the new year info
    lengths['s_year'] = s_year
    lengths['e_year'] = e_year
    return lengths


def detect_resolution(path):
    """Attempt to detect the domain information from the path.

    Args:
        path (str): Path.
    
    Returns:
        int : Domain
    
    Raises:
        ValueError : When the domain can't be detected.
    """
    matches = re.findall(r'[0-9]*km', path)

    num_matches = len(matches)
    if num_matches != 1:
        raise ValueError(f'Unable to detect domain in {path}. There should be exactly 1 domain segment in the path, found {num_matches}')
    
    return int(matches[0].replace('km', ''))
    


# if __name__ == '__main__':
def test():

    # Load configuration
    drs_config = load_package_data('data/drs.json')

    # Parse arguments
    parser = argparse.ArgumentParser()
    parser.description = 'This script post-processes output from CCAM runs into DRS format.'

    parser.add_argument('input_dir', type=str, help='Input directory.')
    parser.add_argument('output_dir', type=str, help='Input directory.')

    # parser.add_argument('odir', type=str, help='Output directory.')
    # parser.add_argument('base', type=str, help='Filename base.')
    parser.add_argument('model', type=str, help='GCM name.')
    parser.add_argument('start_year', type=int, help='Start year.')
    parser.add_argument('end_year', type=int, help='End year.')
    parser.add_argument('ivar', type=str, help='Variable name.')
    parser.add_argument('freq', type=str, help='Output frequency')
    parser.add_argument('project', type=str, help='Funding project.')
    args = parser.parse_args()
    
    # Set up context - starting with the arguments supplied to the script
    # We do things this way because there is a lot of string interpolation
    context = vars(args)

    # Add the defaults
    context.update(drs_config['defaults'])

    # Check if this is a fixed variable
    context['fixed'] = context['ivar'] in drs_config['fixed_variables']
    if context['fixed']:
        context['ensemble'] = 'r0i0p0'

    context['var_in'] = context['ivar']
    context['var_out'] = context['ivar']

    # Get directories and create output location
    context['rundir'] = get_rundir()

    # Create a list of decades to process (usually 1)
    decades = get_decades(args.start_year, args.end_year+1)

    # Detect the resolution from the path information
    context['res_km'] = detect_resolution(args.input_dir)

    # Work out the frequencies
    if context['freq']:
        context['freqs'] = [context['freq']]
        context['no_freq'] = False
    else:
        context['freqs'] = '1hr,day,month'.split(',')
        context['no_freq'] = True

    n_domains = 1

    # Set up the domain information
    domains = get_domains(context['res_km'], args.freq, context['fixed'], context['no_freq'])

    context['run_type'] = 'Climate change'
    context['mode'] = ' bias- and variance-corrected sea surface temperatures'

    # Load the metadata for this project
    meta_model = drs_config['metadata'][args.model]

    meta_model['gcm_model'] = meta_model['gcm_model'] % dict(model=args.model)

    meta_project = drs_config['projects'][args.project]

    # Add the meta to the context, there is some info there that will be needed.
    context.update(meta_model)
    context.update(meta_project)

    # TODO: This could be replaced with a better set of defaults to save conditionals
    # ~L297 in procvar4
    if context['freq'] == 'cordex':

        meta_project['project_combined'] = 'CORDEX'

    elif meta_model['gcm_model'] == 'ERAINT':

        meta_project['project_long'] = "2018 Australian Wine Industry Projections and 2019 DELWP Victorian Climate Projections"
        meta_project['plural'] = 's'
        meta_project['project_combined'] = 'WINE / DELWP'

    else:

        meta_project['plural'] = 's'
        meta_project['project_combined'] = args.project

    # Set starting defaults for start/end years
    context['s_year'] = context['start_year']
    context['e_year'] = context['end_year']

    for decade in decades:

        for file_freq in context['freqs']:
            
            # Update the path freq
            if file_freq == 'fixed':
                path_freq = 'fx'
            elif file_freq == 'month':
                path_freq = 'mon'
            else:
                path_freq = file_freq

            context['path_freq'] = path_freq

            pprint(context)
            sys.exit()
                
            # for domain in domains:


    # # Loop through each decade
    # for decade in decades:

    #     lengths, s_year, e_year = get_lengths_new(decade, **context)

    #     # Update with what has been calculated
    #     context.update(lengths)

    #     # Set these in the context
    #     context['s_year'] = s_year
    #     context['e_year'] = e_year

    #     # Change the formatting depending on the frequency
    #     if freq
    #     start_date = datetime(s_year, 1, 1).strftime('%')
    #     end_date = datetime(e_year, 1, 1)


    #     # Override for 2006
    #     if context['s_year'] < 2006:
    #         context['experiment'] = 'historical'
    #     else:
    #         context['experiment'] = context['rcp']
        
    #     # Override for ERA
    #     if context['gcm_model'] in ['ERAINT', 'ERA5']:
    #         context['experiment'] = 'evaluation'
    #         context['description'] = get_template(drs_config, "description_era") % context
    #     else:
    #         context['description'] = get_template(drs_config, "description_other") % context

    #     # Loop through each domain requested
    #     for domain in domains['domains']:

    #         # Load the domain info from the config (dx etc.)
    #         domain_info = drs_config['domains'][domain]
    #         context['domain'] = domain
    #         context.update(domain_info)

    #         # Special case
    #         if domain == 'AUS-44i':
                
    #             drs_project = 'CORDEX'
                
    #             if file_freq == 'month':
    #                 start_mmdd ='01'
    #                 end_mmdd = '12'

    #         # Interpolate the filename information
    #         if context['fixed']:
    #             context['filename_template'] = get_template(drs_config, 'filename_fixed')
    #         else:
    #             context['filename_template'] = get_template(drs_config, 'filename')

    #         # Build output path
    #         context['drs_base'] = get_template(drs_config, 'drs_base') % context
    #         output_path_template = get_template(drs_config, 'drs_path')
    #         output_path = output_path_template % context
    #         context['output_path_template'] = output_path_template
    #         context['output_path'] = output_path

    #         # Build output filename and put them together
    #         output_filename = get_template(drs_config, 'filename') % context
    #         output_filepath = os.path.join(output_path, output_filename)

    #         # TODO: List of global attributes to retain?
    #         pprint(context)
    #         sys.exit()

    #         # Open the work file
    #         ds = xr.open_dataset(input_filename)

    #         # Find the date the model was run
    #         creation_date = ds.attrs['date_header']

    #         # Get the number of x-y and z grid points of the model grid
    #         il = ds.attrs['il']
    #         kl = ds.attrs['kl']
    #         rlon = ds.attrs['rlong0']
    #         rlat = ds.attrs['rlat0']

    #         # Get grid stretching value of model grid
    #         schmidt = ds.attrs['schmidt']

    #         # Set created date
    #         created = datetime.utcnow()
    #         metadata_mod_date = datetime.utcnow()

    #         # Create a unique identifier for this file
    #         uuid = uuid.uuid4()

    #         # Select domain
    #         # ds = ds.sel()

    #         # Get dimension information
    #         geospatial_lon_min = ds.lon.min()
    #         geospatial_lon_max = ds.lon.max()
    #         geospatial_lat_min = ds.lat.min()
    #         geospatial_lat_max = ds.lat.max()

    #         # Get vertical for 3D data
    #         if 'lev' in ds.coords.keys():
    #             geospatial_vertical_min = ds.lev.min()
    #             geospatial_vertical_max = ds.lev.max()
    #             geospatial_vertical_units = ds.lev.attrs['units']
            
    #         # Skip if dry run
    #         if dry_run:
    #             continue

    #         # Strip global attributes
    #         ds.attrs = dict()

    #         # Read the expected variable metadata out of the spec
    #         # TODO: Allow day or month here (also fixed)
    #         metadata_spec = 'cordex-day.csv'
    #         metadata = load_package_data(f'specifications/{metadata_spec}')
    #         metadata = metadata['variables'][args.ivar]

    #         metadata_updates = dict()

    #         # Extract the 
    #         for attr, schema in metadata.items():
    #             metadata_updates[attr] = schema['allowed'][0]
                
    #         ds[ivar].attrs.update(metadata_updates)

    #         # TODO: Check that the file is of the correct length
    #         num_times = ds.time.shape[0]
            
    #         if num_times != expected_times:
    #             print('ERROR: Newly created file is of wrong, nonzero length')
    #             print(f'Expected {expected_times}, found {num_times}')
            
    #         # Construct the global metadata from all of the logic above
    #         meta_global = dict()

    #         # Put the metadata together from different sources

    #         for meta in [model_metadata, project_metadata]:

    #             # Then individual attributes
    #             for key, value in meta.items():

    #                 # Interpolate the context into the values and assign to the final
    #                 meta_global[key] = value % context

    #         # Assign the metadata to the file
    #         ds.attrs.update(meta_global)

    #         # TODO: Convert to netcdf4 classic and save
    #         os.makedirs(output_path, exist_ok=True)
    #         ds.to_netcdf(output_filepath, format='NETCDF4_CLASSIC')

def get_model_meta(config, model, is_cordex=False):
    """Get the model metadata out of the config.

    Args:
        config (dict): Configuration dictionary.
        model (str): Model name
        is_cordex (bool, optional): Apply cordex overrides. Defaults to False.

    Returns:
        dict: Model metadata.
    """

    meta = config['models'][model]

    # Override with cordex alternatives
    if is_cordex and '_cordex' in meta.keys():
        meta.update(meta['_cordex'])
        meta.pop('_cordex', None)

    return meta


def get_project_meta(config, project, is_cordex=False, is_eraint=False):

    meta = config['projects'][project]

    if is_cordex:

        meta['project_combined'] = 'CORDEX'

    else:

        if is_eraint:

            meta['project_long'] = "2018 Australian Wine Industry Projections and 2019 DELWP Victorian Climate Projections"
            meta['plural'] = 's'
            meta['project_combined'] = 'WINE / DELWP'

        else:

            meta['plural'] = 's'
            meta['project_combined'] = project
    
    return meta

if __name__ == '__main__':

    config = load_package_data('data/drs.json')

    parser = argparse.ArgumentParser()
    parser.description = 'DRS processing script'
    parser.add_argument('input_dir')
    parser.add_argument('output_dir')
    parser.add_argument('-s', '--start_year', required=True, type=int)
    parser.add_argument('-e', '--end_year', required=True, type=int)
    parser.add_argument('-v', '--variable', required=True, type=str)
    parser.add_argument('-f', '--frequency', type=str)
    parser.add_argument('-p', '--project', required=True, type=str)
    parser.add_argument('-m', '--model', required=True, type=str)
    parser.add_argument('-b', '--base', required=True, type=str)
    
    # Initialise context, add defaults
    context = config['defaults']
    context.update(vars(parser.parse_args()))

    # Check if the variable is fixed
    is_fixed = context['variable'] in config['fixed_variables']

    # Get a list of decades
    decades = get_decades(context['start_year'], context['end_year'])
    
    # Detect resolution
    res_km = detect_resolution(context['input_dir'])
    context['res_km'] = res_km

    # Get frequencies
    no_frequency = context['frequency'] is None
    is_cordex = context['frequency'] == 'cordex'

    if no_frequency:
        frequencies = '1hr,day,month'.split(',')
    else:
        frequencies = [context['frequency']]

    context['frequencies'] = frequencies

    # Get domain information
    domain_info = get_domains(
        res_km,
        context['frequency'],
        is_fixed,
        no_frequency
    )

    context.update(domain_info)
    domains = domain_info['domains']

    # Get model information
    meta_model = get_model_meta(config, context['model'])
    context.update(meta_model)
    is_eraint = meta_model['gcm_model'] == 'ERAINT'

    # Get project information
    meta_project = get_project_meta(
        config,
        context['project'],
        is_cordex=is_cordex,
        is_eraint=is_eraint   
    )

    context.update(meta_project)

    context['s_year'] = context['start_year']
    context['e_year'] = context['end_year']

    for decade in decades:

        # Update length information
        length_info = get_lengths_new(decade, **context)
        context.update(length_info)

        # TODO: Need a cleaner way to do this
        if context['s_year'] < 2006:
            context['experiment'] = 'historical'
        else:
            context['experiment'] = context['rcp']
        
        if context['gcm_model'] in ['ERAINT', 'ERA5']:
            context['experiment'] = 'evaluation'
            description_template = 'description_era'
        else:
            description_template = 'description_other'

        context['description'] = get_template(config, description_template) % context

        pprint(context)
        sys.exit()

        for frequency in frequencies:
            for domain in domains:
                print(decade, frequency, domain)

    # if context['freq'] == 'cordex':

    #     meta_project['project_combined'] = 'CORDEX'

    # elif meta_model['gcm_model'] == 'ERAINT':

    #     meta_project['project_long'] = "2018 Australian Wine Industry Projections and 2019 DELWP Victorian Climate Projections"
    #     meta_project['plural'] = 's'
    #     meta_project['project_combined'] = 'WINE / DELWP'

    # else:

    #     meta_project['plural'] = 's'
    #     meta_project['project_combined'] = args.project

    
    pprint(context)


from axiom.drs.payload import generate_payloads

def test_generate_payloads():
    """Test generate_payloads."""

    payloads = generate_payloads('files.nc', 'output', 2000, 2021, 'CORDEX', 'ACCESS', 'AUS-10i', ['tas', 'tasmax'], 'CORDEX', ['1H', '6H'], 1, test_meta='test_meta_value')

    p1 = payloads[0]
    p2 = payloads[-1]

    assert len(payloads) == 44 # 22years x 2 time freq
    assert p1.start_year == 2000 and p1.project == 'CORDEX'
    assert p2.start_year == 2021 and p2.model == 'ACCESS'





"""Test utilities."""
import axiom.utilities as au
import xarray as xr
import numpy as np
import pytest


def test_isolate_coordinate():
    """Test the isolate_coordinate function."""

    # Create an object with time, lat, lon
    obj = xr.DataArray(
        data=np.arange(54).reshape(6, 3, 3),
        coords={'time': [1, 2, 3, 4, 5, 6], 'lat': [1, 2, 3], 'lon': [1, 2, 3]},
        dims=['time', 'lat', 'lon'],
    )

    # Isolate the lat coordinate
    result = list(au.isolate_coordinate(obj, 'lat', drop=True).coords)

    # Ensure we get only one and that is the one we want.
    assert len(result) == 1
    assert result[0] == 'lat'


def test_load_package_json():
    """Test the load_package_json function."""
    # Load something that exists
    result = au.load_package_json('drs.json')
    assert isinstance(result, dict)

    # Test something that does not exist
    with pytest.raises(FileNotFoundError):
        result = au.load_package_json('does_not_exist.json')


"""Custom exception classes."""

class ResolutionDetectionException(Exception):
    """Raised when tere are issues detecting resolution information."""
    pass

class MalformedDRSJSONPayloadException(Exception):
    """Raised when a DRS JSON payload can't be processed."""
    pass

class NoFilesToProcessException(Exception):
    """Raised when there are no files to process for a DRS payload"""
    pass

class DRSContextInterpolationException(Exception):
    """Raised when there are remaining placeholders in a template after DRS context interpolation.
    
    Args:
        placeholders (list) : List of placeholders that have yet to be interpolated.
    """
    def __init__(self, placeholders):
        msg = 'The following placeholders have been unsuccessfully interpolated:\n'
        msg += '\n'.join(placeholders)
        super().__init__(msg)



"""Class to represent a JSON payload."""
import enum
import json
import axiom.schemas as axs
import axiom.utilities as au


class Payload:

    """Class to represent a Payload object.

    Args:
        input_files (str): Globbable path to input files.
        output_directory (str): Output directory.
        start_year (int) : Start year.
        end_year (int): End year.
        output_frequency (str): Output frequency.
        project (str): Project key from projects.json.
        model (str): Model key from models.json.
        domain (str): Domain key from domains.json.
        variables (list, optional): List of variables to process. Defaults to [].
        **kwargs : Additional key/value pairs added as a metadata.
    
    Attributes:
        input_files (str): Globbable path to input files.
        output_directory (str): Destination output directory (DRS built from here).
        project (str): Project key from projects.json.
        model (str): Model key from models.json.
        domain (str): Domain key from domains.json.
        start_year (int): Start year.
        end_year (int): End year.
        variables (list): List of variable names to process.
        output_frequency (str): Output frequency/resolution of data.
        extra (dict): Additional metadata key/value pairs.
    """

    def __init__(self, input_files, output_directory, start_year, end_year, output_frequency, project, model, domain, variables=[], batch=1, **kwargs):
        
        self.input_files = input_files
        self.output_directory = output_directory
        
        self.project = project
        self.model = model
        self.domain = domain

        self.start_year = start_year
        self.end_year = end_year
        
        self.variables = variables
        self.output_frequency = output_frequency

        self.batch = batch

        self.extra = kwargs
    

    def get_filename(self):
        """Generate a filename for this payload.

        Returns:
            str: Filename
        
        Examples:
            >>> filename = payload.get_filename()
        """
        bbb = str(self.batch).zfill(3)
        return f'payload.{self.start_year}.{self.output_frequency}.{bbb}.json'
    

    def to_dict(self):
        """Convert the Payload object to a dictionary.

        Returns:
            dict: Dictionary representation of the Payload.
        
        Examples:
            >>> payload.to_dict()
        """
        _dict = self.__dict__.copy()
        extra = _dict.pop('extra')
        _dict.update(extra)
        return _dict
    

    def from_dict(d):
        """Create a Payload from a dictionary.

        Args:
            d (dict): Dictionary.

        Returns:
            axiom.drs.Payload: Payload object.
        
        Examples:
            >>> payload = Payload.from_dict(d)
        """
        return Payload(**d)
    

    def to_json(self, filepath=None):
        """Serialize the Payload to the filepath (save as JSON).

        Args:
            filepath (str, optional): Filepath to write. Defaults to None (Writes default filename to current directory).
        
        Examples:
            >>> payload.to_json('payload.json')
        """
        if filepath is None:
            filepath = self.get_filename()

        _dict = self.to_dict()
        with open(filepath, 'w') as f:
            f.write(json.dumps(_dict, indent=4))


    def from_json(filepath):
        """Instantiate a payload from a json filepath.

        Args:
            filepath (str): Path.

        Returns:
            axiom.drs.Payload: Payload.
        
        Examples:
            >>> payload = Payload.from_json('payload.json')
        """
        d = json.load(open(filepath, 'r'))
        return Payload.from_dict(d)


def generate_payloads(input_files, output_directory, start_year, end_year, project, model, domain, variables=None, schema=None, output_frequencies=['1H', '6H', '1D', '1M'], num_batches=1, **extra):
    """Generate payload files.

    Args:
        input_files (str): Globbable path to input files.
        output_directory (str): Destination filepath, DRS will be built from here.
        start_year (int): Start year.
        end_year (int): End year.
        project (str): Project key from projects.json.
        model (str): Model key from models.json.
        domain (str): Domain key from domains.json.
        variables (list(str), optional): List of variables to project. Defaults to None (read all from schema).
        schema (str): Schema name or filepath.
        output_frequencies (list(str), optional): List of output frequencies. Defaults to ['1H', '6H', '1D', '1M'].
        num_batches (int, optional): Number of batches to split processing into. Defaults to 1.
        **extra : Key/value pairs added as additional metadata.
    
    Returns:
        list : A list of payload objects.
    """

    payloads = list()

    # Load the variables from the schema if not provided
    if variables is None:
        schema = axs.load_schema(schema)
        variables = list(schema['variables'].keys())    
    
    # Batch if required (could be a single batch)
    batches = au.batch_split(variables, num_batches)

    for year in range(start_year, end_year+1):

        for output_frequency in output_frequencies:

            for batch_ix, batch in enumerate(batches):
        
                # Generate a payload
                payload = Payload(
                    input_files=input_files,
                    output_directory=output_directory,
                    project=project,
                    model=model,
                    domain=domain,
                    start_year=year,
                    end_year=year,
                    variables=batch,
                    output_frequency=output_frequency,
                    batch=batch_ix,
                    **extra
                )

                payloads.append(payload)
    
    return payloads


"""Tests for decorator methods."""
from axiom.decorators import metadata
import xarray as xr


def test_metadata():
    """Test the metadata decorator, which adds metadata to the return value of a function."""
    data = [1, 2, 3, 4]
    da = xr.DataArray(data)

    @metadata(name='temp', units='K')
    def _func(da):
        return da
    
    da = _func(da)

    assert da.attrs['units'] == 'K' and da.name == 'temp'


class Domain:

    """Domain specification class.

    Args:
        name (str): Name of the domain.
        dx (float): Delta-x.
        lat_min (float): Minimum latitude.
        lat_max (float): Maximum latitude.
        lon_min (float): Minimum longitude.
        lon_max (float): Maximum longitude.
    """

    def __init__(self, name, dx, lat_min, lat_max, lon_min, lon_max):
        self.name = name
        self.dx = dx
        self.lat_min = lat_min
        self.lat_max = lat_max
        self.lon_min = lon_min
        self.lon_max = lon_max
    

    def from_dict(domain_dict):
        """Convert from a dictionary definition to a Domain class object.
        
        Args:
            domain_dict (dict) : Domain dictionary.
        
        Returns:
            Domain : Domain class.
        """
        return Domain(**domain_dict)
    

    def to_dict(self):
        """Convert the internal representation into a dict object.

        Returns:
            dict : Dictionary.
        """
        return dict(
            name=self.name,
            dx=self.dx,
            lat_min=self.lat_min,
            lat_max=self.lat_max,
            lon_min=self.lon_min,
            lon_max=self.lon_max
        )
    

    def from_directive(directive):
        """Generate domain object from a directive.

        Example: 'Antarctica,1.0,-180.0,180.0-90.0,-50.0'

        Args:
            directive (str): Domain directive.
        
        Returns:
            Domain : Domain class.
        """
        name, *directives = directive.split(',')
        dx, lat_min, lat_max, lon_min, lon_max = map(float, directives)
        return Domain(
            name=name,
            dx=dx,
            lat_min=lat_min,
            lat_max=lat_max,
            lon_min=lon_min,
            lon_max=lon_max
        )
    

    def to_directive(self):
        """Convert the internal representation into a directive string.

        Returns:
            str : Directive.
        """
        return f'{self.name},{self.dx},{self.lat_min},{self.lat_max},{self.lon_min},{self.lon_max}'


    def subset_xarray(self, ds, drop=True):
        """Subset an xarray object with this domain object.

        Args:
            ds (xarray.Dataset or xarray.DataArray)
            drop (bool) : Drop the data outside of the domain. Defaults to True.
        
        Returns:
            xarray.Dataset or xarray.DataArray : Object subset with this domain.
        """

        lon_min, lon_max = self.lon_min, self.lon_max

        # Fix to cross the meridian
        if lon_max < lon_min:
            lon_min, lon_max = lon_max, lon_min

        return ds.sel(lat=slice(self.lat_min, self.lat_max), lon=slice(lon_min, lon_max))
    

    def from_config(key, config):
        """Generate a domain object a section in a parsed configparser config object.

        Args:
            key (str): Key of the domain in the section.
            config (configparser.Config) : Configuration object.
        """
        # Get the section
        elements = {key: float(value) for key, value in config[key].items()}
        return Domain(
            name=key,
            **elements
        )


"""Main entrypoint for DRS processing."""
from genericpath import isfile
import os
import argparse
from datetime import datetime
from uuid import uuid4
import xarray as xr
import axiom.utilities as au
import axiom.drs.utilities as adu
from axiom.drs.domain import Domain
import axiom.schemas as axs
import json
import sys
from distributed import Client, LocalCluster
from axiom.config import load_config
from axiom import __version__ as axiom_version
from axiom.exceptions import NoFilesToProcessException, DRSContextInterpolationException
import shutil
from dask.distributed import progress, wait
import numpy as np
from axiom.supervisor import Supervisor


def consume(json_filepath):
    """Consume a json payload (for message passing)

    Args:
        json_filepath (str): Path to the JSON file.
    """
    logger = au.get_logger(__name__)

    # Check if the file has already been consumed
    consumed_filepath = json_filepath.replace('.json', '.consumed')
    if os.path.isfile(consumed_filepath):
        logger.info(
            f'{json_filepath} has already been consumed and needs to be cleaned up by another process. Terminating.')
        sys.exit()

    # Check if the file is locked
    if au.is_locked(json_filepath):
        logger.info(
            f'{json_filepath} is locked, possibly by another process. Terminating.')
        sys.exit()

    # Lock the file
    au.lock(json_filepath)

    # Convert to dict
    payload = json.loads(open(json_filepath, 'r').read())

    # Allow rerun of failed variables (do this after all other variables have been processed!)
    config = load_config('drs')
    failures_path = f'{json_filepath}_001.failed'
    if config.rerun_failures and os.path.exists(failures_path):
        failed_variables = open(failures_path, 'r').read().splitlines()
        payload['variables'] = failed_variables

    # Process
    process_multi(**payload)

    # Mark consumed by touching another file.
    au.touch(consumed_filepath)

    # Unlock
    au.unlock(json_filepath)

    # Explicit exit (#125)
    sys.exit(0)


def process(
    input_files,
    output_directory,
    variable,
    project,
    model,
    domain,
    start_year, end_year,
    output_frequency,
    level=None,
    input_resolution=None,
    overwrite=True,
    preprocessor=None,
    postprocessor=None,
    **kwargs
):
    """Method to process a single variable/domain/resolution combination.

    Args:
        input_files (str or list): Globbable string or list of filepaths.
        output_directory (str) : Path from which to build DRS structure.
        variable (str): Variable to process.
        project (str): Project metadata to apply (loaded from user config).
        model (str): Model metadata to apply (loaded from user config).
        start_year (int): Start year.
        end_year (int): End year.
        output_frequency (str): Output frequency to process.
        input_resolution (float, optional): Input resolution in km. Leave black to auto-detect from filepaths.
        overwrite (bool): Overwrite the data at the destination. Defaults to True.
        preprocessor (str): Data preprocessor to activate on input data. Defaults to None.
        postprocesser (str): Data postprocess to activate before writing data. Defaults to None.
        **kwargs: Additional keyword arguments used in metadata interpolation.
    """

    # Start the clock
    timer = au.Timer()
    timer.start()

    # Capture what was passed into this method for interpolation context later.
    local_args = locals()

    # Load the logger and configuration
    logger = au.get_logger(__name__)
    config = load_config('drs')

    # Dump the job id if available
    if 'PBS_JOBID' in os.environ.keys():
        jobid = os.getenv('PBS_JOBID')
        logger.info(f'My PBS_JOBID is {jobid}')

    logger.info('Searching for files matching the following path:')
    logger.info(input_files)

    # Get a list of the potential filepaths
    input_files = au.auto_glob(input_files)
    num_files = len(input_files)
    logger.debug(f'{num_files} to consider before filtering.')

    # Filter by those that actually have the variable in the filename.
    if config.filename_filtering['variable']:

        input_files = adu.filter_by_variable_name(input_files, variable)
        num_files = len(input_files)
        logger.debug(
            f'{num_files} to consider after filename variable filtering.')

    # Filter by those that actually have the year in the filename (plus or minus an offset).
    if config.filename_filtering['year']:
        
        input_files = filter_years(
            input_files, start_year, offset=config.filename_filtering['year_offset'])
        num_files = len(input_files)
        logger.debug(f'{num_files} to consider after filename year filtering.')

    # Is there anything left to process?
    if len(input_files) == 0:
        raise NoFilesToProcessException()

    # Dump filepaths prior to loading
    if config.get('dump_filepaths_prior_to_loading', default=False):
        logger.debug('Dumping filepaths prior to loading... there may be a lot.')
        for input_file in input_files:
            logger.debug(input_file)

    # Detect the input resolution if it it not supplied
    if input_resolution is None:
        logger.debug('No input resolution supplied, auto-detecting')
        input_resolution = adu.detect_resolution(input_files)
        logger.debug(f'Input resolution detected as {input_resolution} km')

    # Load project config
    logger.info(f'Loading project config ({project})')
    project_key = project
    project = load_config('projects')[project_key]

    # Ensure it actually exists
    assert isinstance(project, dict), f'Project {project_key} not found, does it exist in projects.json?'

    # Load model config
    logger.info(f'Loading model config ({model})')
    model_key = model
    model = load_config('models')[model_key]

    # Ensure it actually exists
    assert isinstance(model, dict), f'Model {model_key} not found, does it exist in models.json?'

    logger.debug(
        'Loading files into distributed memory, this may take some time.')

    # TODO: Remove!!!! This is just to make CCAM work in the short term
    if preprocessor is None and 'ccam' in input_files[0] and config.auto_detect_ccam == True:
        logger.warn('CCAM preprocessor override used')
        preprocessor = 'ccam'
        postprocessor = 'ccam'

    # Load a preprocessor, if one exists.
    preprocessor = adu.load_preprocessor(preprocessor)
    def preprocess(ds, *args, **kwargs): return preprocessor(ds, **local_args)

    # Load the open_dataset configuration
    open_dataset_kwargs = config['xarray']['open_dataset']

    # Account for fixed variables, if defined
    if 'variables_fixed' in project.keys() and variable in project['variables_fixed']:

        # Load just the first file
        ds = xr.open_dataset(input_files[0], **open_dataset_kwargs)
        ds = preprocess(ds, variable=variable)

    else:
        
        ds = xr.open_mfdataset(
            input_files,
            preprocess=preprocess,
            **open_dataset_kwargs
        )

    # Subset temporally
    if not adu.is_time_invariant(ds):
        logger.info(f'Subsetting times to {start_year}')
        # ixs = np.where(ds['time.year'] == start_year)
        time_slice = slice(f'{start_year}-01-01', f'{start_year}-12-31')
        # ds = ds.isel(time=slice(ixs[0][0], ixs[0][-1] + 1))
        ds = ds.sel(time=time_slice, drop=True)

    # Skip over the file if subdaily resampling is disabled, this will stop 
    native_frequency = adu.detect_input_frequency(ds)

    # Ensure blank output frequency is indeed fixed and only one can be written
    if adu.is_time_invariant(ds):
        output_frequency = 'fx'
        overwrite = False

    logger.info(f'native_frequency = {native_frequency}, output_frequency = {output_frequency}')
    if config.allow_subdaily_resampling == False and native_frequency != output_frequency and 'H' in output_frequency:
        logger.info(f'Subdaily resampling has been disabled and input/output frequencies do not match, skipping {variable}.')
        return

    # Persist now, get it on the cluster while the rest of the metadata assembly continues
    ds = ds.persist()

    # Determine time-invariance
    time_invariant = 'time' not in list(ds.coords.keys())

    # Assemble the context object (order dependent!)
    logger.debug('Assembling interpolation context.')
    context = config.metadata_defaults.copy()

    # Add metadata from the input data
    context.update(ds.attrs)

    # Add user-supplied metadata
    context.update(kwargs)

    # Add project and model metadata
    context.update(project)
    context.update(model)

    # Add additional args
    context.update(local_args)
    context['res_km'] = input_resolution

    # Sort the dimensions (fixes domain subsetting)
    logger.debug('Sorting data')
    sort_coords = list()
    for coord in 'time,lev,lat,lon'.split(','):
        if coord in ds.coords.keys():
            sort_coords.append(coord)

    ds = ds.sortby(sort_coords)

    logger.debug('Applying metadata schema')

    # Load a user-supplied schema, if one exists.
    if 'schema' in kwargs.keys():
        schema_key = kwargs['schema']
    else:
        schema_key = config['default_schema']

    schema = axs.load_schema(schema_key)
    ds = au.apply_schema(ds, schema)

    logger.info(f'Parsing domain {domain}')
    if isinstance(domain, str):

        # Registered domain
        if adu.is_registered_domain(domain):
            domain = adu.get_domain(domain)

        # Attempt to parse
        else:
            domain = Domain.from_directive(domain)

    # We will only otherwise accept a domain object.
    elif isinstance(domain, Domain) == False:
        raise Exception(f'Unable to parse domain {domain}.')

    logger.debug('Domain: ' + domain.to_directive())

    # Subset the geographical domain
    logger.debug('Subsetting geographical domain.')
    ds = domain.subset_xarray(ds, drop=True)

    # TODO: Need to find a less manual way to do this.
    for year in adu.generate_years_list(start_year, end_year):

        logger.info(f'Processing {year}')

        # Subset the data into just this year
        if not time_invariant:
            time_slice = slice(f'{year}-01-01', f'{year}-12-31')
            _ds = ds.sel(time=time_slice, drop=True)
        else:
            _ds = ds.copy()

        # Historical cutoff is defined in $HOME/.axiom/drs.json
        if config.enable_historical_cutoff == True:
            context['experiment'] = 'historical' if year < config.historical_cutoff else context['rcp']
        
        logger.info(f'Native frequency of data detected as {native_frequency}')

        # Automatically detect the output_frequency from the input data, this will not require resampling
        
        # Flag to trigger cell_method update below.
        resampling_applied = False

        if output_frequency == 'from_input' or output_frequency == native_frequency:
            output_frequency = adu.detect_input_frequency(_ds)
            logger.info(
                f'output_frequency detected from inputs ({output_frequency})')
            logger.info(f'No need to resample.')
            # Map the frequency to something DRS-compliant
            context['frequency_mapping'] = config['frequency_mapping'][output_frequency]

        # Fixed variables, just change the frequency_mapping
        elif adu.is_time_invariant(_ds):
            output_frequency = 'fx'
            logger.info(
                'Data is time-invariant (fixed variable), overriding frequency_mapping to fx')
            context['frequency_mapping'] = 'fx'

        # Actually perform the resample
        else:
            logger.debug(f'Resampling to {output_frequency} mean.')
            context['frequency_mapping'] = config['frequency_mapping'][output_frequency]
            _ds = _ds.resample(time=output_frequency, label='left').mean()

            # Update the cell methods below
            resampling_applied = True

        # Start persisting the computation now
        _ds = _ds.persist()

        # Monthly data should have the days truncated
        # context['start_date'] = f'{year}0101' if output_frequency[-1] != 'M' else f'{year}01'
        # context['end_date'] = f'{year}1231' if output_frequency[-1] != 'M' else f'{year}12'

        context['start_date'], context['end_date'] = adu.get_start_and_end_dates(year, output_frequency)

        # Tracking info
        context['creation_date'] = datetime.utcnow()
        context['uuid'] = uuid4()

        # Interpolate context
        logger.info('Interpolating context.')
        context = adu.interpolate_context(context)

        # Assemble the global meta, add axiom details
        logger.debug('Assembling global metadata.')
        global_attrs = dict(
            axiom_version=axiom_version,
            axiom_schema=schema_key
        )

        for key, value in config.metadata_defaults.items():
            global_attrs[key] = str(value) % context

        # Strip and reapply metadata
        logger.debug('Applying metadata')
        _ds.attrs = global_attrs

        # Add in the variable to the context
        context['variable'] = variable

        # Reapply the schema
        logger.info('Reapplying schema')
        _ds = au.apply_schema(_ds, schema)

        # Copy coordinate attributes straight off the inputs
        if config.copy_coordinates_from_inputs:
            for coord in list(_ds.coords.keys()):
                _ds[coord].attrs = ds[coord].attrs

        # Assemble the encoding dictionaries (to ensure time units work!)
        logger.debug('Applying encoding')
        encoding = dict()

        for coord in list(_ds.coords.keys()):
            if coord not in config.encoding.keys():
                logger.warn(
                    f'Coordinate {coord} is not specified in drs.json file, omitting encoding.')
                continue
            encoding[coord] = config.encoding[coord]

        # Apply a blanket variable encoding.
        encoding[variable] = config.encoding['variables']

        # Postprocess data if required
        postprocessor = adu.load_postprocessor(postprocessor)

        def postprocess(_ds, *args, **kwargs):
            combined = dict()
            combined.update(kwargs)
            combined.update(local_args)
            combined['resampling_applied'] = resampling_applied
    
            return postprocessor(_ds, **combined)
        
        _ds = postprocess(_ds)

        # Update the cell methods
        if resampling_applied:
            _ds = update_cell_methods(_ds, variable, dim='time', method='mean')

        # Get the full output filepath with string interpolation
        logger.debug('Working out output paths')

        # Derive the start/end date strings from the actual timeseries and override
        if config.derive_filename_times_from_data:
            logger.info(
                'User has requested that filename times reflect the actual timeseries.')
            str_times = _ds.time.dt.strftime('%Y%m%d').data
            context['start_date'] = str_times[0]
            context['end_date'] = str_times[-1]
            logger.debug(
                'start_date = %(start_date)s, end_date = %(end_date)s' % context)

        drs_path = adu.get_template(config, 'drs_path') % context
        filename_template = adu.get_template(config, 'filename')

        # Override for fixed variables
        if adu.is_time_invariant(_ds):
            logger.debug('Overriding output filename template with fixed alternative.')
            filename_template = adu.get_template(config, 'filename_fixed')

        # Assemble the output filepath
        output_filename = filename_template % context
        output_filepath = os.path.join(
            output_directory, drs_path, output_filename)
        logger.debug(f'output_filepath = {output_filepath}')

        # Skip if already there and overwrite is not set, otherwise continue
        if os.path.isfile(output_filepath) and overwrite == False:
            logger.debug(
                f'{output_filepath} exists and overwrite is set to False, skipping.')
            continue

        # Check for uninterpolated keys in the output path, which should fail at this point.
        uninterpolated_keys = adu.get_uninterpolated_placeholders(
            output_filepath)

        if len(uninterpolated_keys) > 0:
            logger.error('Uninterpolated keys remain in the output filepath.')
            logger.error(f'output_filepath = {output_filepath}')
            raise DRSContextInterpolationException(uninterpolated_keys)

        # Create the output directory
        output_dir = os.path.dirname(output_filepath)
        logger.debug(f'Creating {output_dir}')
        os.makedirs(output_dir, exist_ok=True)

        # Supervise this job to ensure that it does in fact complete.
        with Supervisor(seconds=config.processing_timeout_seconds, error_msg=f'Variable {variable} took too long to complete, moving on.'):
            logger.info('Waiting for computations to finish.')
            progress(_ds)

        logger.debug(f'Writing {output_filepath}')
        write = _ds.to_netcdf(
            output_filepath,
            format='NETCDF4',
            encoding=encoding,
            unlimited_dims=['time']
        )

    elapsed_time = timer.stop()
    logger.info(f'DRS processing task took {elapsed_time} seconds.')


def load_variable_config(project_config):
    """Extract the variable configuration out of the project configuration.

    Args:
        project_config (dict-like): Project configuration.

    Returns:
        dict: Variable dictionary with name: [levels] (single level will have a list containing None.)
    """

    # Extract the different rank variables
    v2ds = project_config['variables_2d']
    v3ds = project_config['variables_3d']

    # Create a dictionary of variables to process keyed to an empty list of levels for 2D
    variables = {v2d: [None] for v2d in v2ds}

    # Add in the 3D variables, with levels this time
    for v3d, levels in v3ds.items():
        variables[v3d] = levels

    return variables


def process_multi(variables, domain, project, **kwargs):
    """Start a processing chain of multiple variables.

    Args:
        variables (list): List of variables to process.
        domain (str): Domain to process from domains.json.
        project (str): Project metadata to use from projects.json.
        **kwargs: Additional keyword arguments to pass to the processing chain.
    """

    logger = au.get_logger(__name__)

    # Load the project metadata
    project_config = load_config('projects')[project]
    config = load_config('drs')

    # Load all variables if nothing was supplied
    if not variables:

        # Select a default schema
        schema_key = config['default_schema']

        # Override if requested, this might be a direct filepath
        if 'schema_key' in kwargs.keys():
            schema_key = kwargs['schema']

        # Load it
        logger.info(f'No variables supplied, loading from schema as defined in configuration ({schema_key}).')
        schema = axs.load_schema(schema_key)        
        variables = list(schema['variables'].keys())
    
    
    else:
        logger.debug('User has supplied the following variables')
        logger.debug(variables)

    num_variables = len(variables)
    logger.info(f'{num_variables} variable(s) to process.')

    # Start the cluster if requested
    if config.dask['enable']:
        logger.info('Starting dask client.')

        cluster_config = config.dask['cluster']

        # Add PBS_JOBFS if set
        if 'PBS_JOBFS' in os.environ.keys():
            cluster_config['local_directory'] = os.getenv('PBS_JOBFS')

        cluster = LocalCluster(**cluster_config)
        client = Client(cluster)
        logger.info(client)

    output_frequencies = au.pluralise(kwargs['output_frequency'])
    
    # Yes this is a nested loop, but a single variable/domain/output_freq combination could still be 10K+ files, which WILL be processed in parallel.
    for variable in variables:
        for output_frequency in output_frequencies:

            attempt = 1
            no_files = False
            success = False

            rerun_attempts = config.get('rerun_attempts', default=1)

            while attempt <= rerun_attempts and no_files == False and success == False:

                logger.info(f'Processing {variable} {output_frequency}')
                instance_kwargs = kwargs.copy()
                instance_kwargs['variable'] = variable
                instance_kwargs['domain'] = domain
                instance_kwargs['project'] = project
                instance_kwargs['output_frequency'] = output_frequency

                if config.dask['enable']:
                    logger.info('Waiting for dask workers')
                    client.wait_for_workers(1, timeout=config['dask']['restart_timeout_seconds'])
                    logger.info(client)
                    logger.info(f'Dashboard located at {client.dashboard_link}')

                try:

                    process(**instance_kwargs)
                    success = True

                # Not technically an error, filtering has discounted all available files.
                except NoFilesToProcessException as ex:

                    logger.info(f'No files to process for {variable}')
                    no_files = True

                # Something wrong with the inputs regarding time.
                except IndexError as ex:

                    # Log the exception
                    log_exception(f'Timeseries inconsistency, check input data.', ex)

                    # Check recoverability
                    if is_error_recoverable(ex) and attempt <= rerun_attempts:
                        logger.info('Error is recoverable, incrementing attempts.')
                        attempt += 1
                        continue
                    
                    # Track the failure and max out the attempts to execute the finally clause
                    track_failure(variable, ex)
                    attempt = rerun_attempts + 1

                # Unknown exception
                except Exception as ex:
                    
                    log_exception(
                        f'Variable {variable} failed for output_frequency {output_frequency}. Error to follow',
                        ex
                    )

                    # Check recoverability
                    if is_error_recoverable(ex) and attempt <= rerun_attempts:
                        logger.info(
                            'Error is recoverable, incrementing attempts.')
                        attempt += 1
                        continue

                    # Track the failure and max out the attempts to execute the finally clause
                    track_failure(variable, ex)
                    attempt = rerun_attempts + 1

                # Run regardless of success/failure
                finally:
                    
                    if config.dask['enable'] and config.dask['restart_client_between_variables'] and no_files == False:
                        logger.info('User has requested dask client restarts between each variable (for resilience), restarting now.')
                        client.restart()
                        logger.info(client)

    logger.info('DRS processing complete, please see consumed/lock/log files for further detail.')


def filter_years(filepaths, year, offset=0):
    """Filter filepaths based on a year, plus or minus an offset.

    Args:
        filepaths (list): List of filepaths.
        year (int): Year.
        offset (int, optional): Number of years either side of YEAR to include. Defaults to 0.
    
    Returns:
        list : List of filtered filepaths.
    """
    _filepaths = list()
    years = range(year-offset, year+offset+1)
    for filepath in filepaths:
        for year in years:
            if str(year) in os.path.basename(filepath):
                _filepaths.append(filepath)
    
    return _filepaths


def update_cell_methods(ds, variable, dim='time', method='mean'):
    """Update the cell_methods attribute on the data.

    Args:
        ds (xarray.Dataset): Data.
        variable (str): Variable being processed.
        dim (str): Dimension over which method was applied.
        method (str): Method that was applied.
    """

    da = ds[variable]

    # If there is no cell_methods attribute, add it now.
    if 'cell_methods' not in da.attrs.keys():
        da.attrs['cell_methods'] = f'{dim}: {method}'

    # If the cell method was point, change it
    elif da.attrs['cell_methods'] == f'{dim}: point':
        da.attrs['cell_methods'] = f'{dim}: {method}'

    # If another operation was already applied and doesn't match this, append
    elif da.attrs['cell_methods'] != f'{dim}: {method}':
        da.attrs['cell_methods'] = da.attrs['cell_methods'] + f' {dim}: {method}'

    ds[variable] = da
    return ds


def log_exception(message, ex):
    """Log the exception.

    Args:
        message (str): Human-readable error message.
        ex (Exception): Stack trace.
    """
    logger = au.get_logger(__name__)
    logger.error(message)
    logger.exception(ex)


def is_error_recoverable(exception):
    """Check if the error is recoverable.

    Args:
        exception (Exception): Raised exception to check.
    
    Returns:
        bool : True if recoverable, False otherwise.
    """
    config = load_config('drs')
    return adu.is_error_recoverable(exception, config.get('recoverable_errors', list()))


def track_failure(variable, exception):
    """Track the failure for reprocessing.

    Args:
        variable (str): Variable name.
        exception (Exception): Exception raised.
    """
    
    config = load_config('drs')

    if config.track_failures and 'AXIOM_LOG_DIR' in os.environ.keys() and 'PBS_JOBNAME' in os.environ.keys():

        failed_filepath = os.path.join(
            os.getenv('AXIOM_LOG_DIR'),
            os.getenv('PBS_JOBNAME') + '.failed'
        )

        exname = type(exception).__name__

        with open(failed_filepath, 'a') as failed:  
            failed.write(f'{variable},{exname}\n')



"""Supervisor class for handling long-running jobs."""
import signal

class Supervisor:

    """Supervisor context manager class. Used to supervise a section of code and ensure it completes in a given time.

    Usage:
        >>> with Supervisor(seconds=10, error_msg='Code took too long.'):
        >>>     # ... code that should complete in less than 10 seconds.

    Args:
        seconds (int): Number of seconds to allow for the contained code to run.
        error_msg (str): Error message to raise with the TimeoutError.
    """
 
    def __init__(self, seconds, error_msg):
        self.seconds = seconds
        self.error_msg = error_msg
    

    def _timeout(self, signum, frame):
        """Raise a TimeoutError.

        Args:
            signum : Not used, added for compatibility.
            frame : Not used, added for compatibility.

        Raises:
            TimeoutError: When the number of seconds expires.
        """
        raise TimeoutError(self.error_msg)


    def __enter__(self):
        """Enter the context manager."""
        # Start a timer for how long the code may be in this manager.
        signal.signal(signal.SIGALRM, self._timeout)
        signal.alarm(self.seconds)
    

    def __exit__(self, type, value, traceback):
        """Exit the context manager."""
        # Cancel any set timers.
        signal.alarm(0)


"""General utilities."""
import logging
import sys
import json
import xml.etree.ElementTree as et
import pandas as pd
from datetime import datetime
import xarray as xr
import pkgutil
from collections import namedtuple
import glob
import os
from configparser import ConfigParser, ExtendedInterpolation
from pathlib import Path
import importlib
import time
import subprocess as sp
import numpy as np
from jinja2 import Environment, BaseLoader


def dict2obj(d):
    """Convert a dictionary to an object."""
    obj = namedtuple("dict2obj", d.keys())(*d.values())
    return obj


def get_logger(name, level='debug'):
    """Get a logging object.

    Args:
        name (str): Name of the module currently logging.
        level (str, optional): Level of logging to emit. Defaults to 'debug'.

    Returns:
        logging.Logger: Logging object.
    """

    logger = logging.Logger(name)
    handler = logging.StreamHandler(sys.stdout)
    level = getattr(logging, level.upper())
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    logger.setLevel(level)
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    return logger


def get_variables_and_coordinates(ds):
    """Get a list of variable and coordinate names.

    Args:
        ds (xarray.Dataset): Dataset

    Returns:
        list: List of variable and coordinate names
    """
    return list(ds.data_vars.keys()) + list(ds.coords.keys())


def in_ds(variable, ds):
    """Test if variable is in the data file.

    Args:
        variable (str): Variable name.
        ds (xarray.Dataset): Data.

    Returns:
        bool : True if the variable exists, False otherwise.
    """
    return variable in get_variables_and_coordinates(ds)


def has_variable(ds, variable):
    """Check if a dataset contains a given variable.

    Args:
        ds (xarray.Dataset): Dataset.
        variable (str): Variable name.

    Returns:
        bool: True if ds contains variable, False otherwise.
    """
    return variable in list(ds.data_vars.keys())


def has_coord(dx, coord):
    """Check if a DataArray or Dataset contains a coordinate.

    Args:
        dx (xarray.DataArray or xarray.Dataset): Data.
        coord (str): Coordinate name.

    Returns:
        bool: True if coordinate in object, False otherwise.
    """
    return coord in list(dx.coords.keys())


def has_dim(dx, dim):
    """Check if a DataArray or Dataset contains a dimension.

    Args:
        dx (xarray.DataArray or xarray.Dataset): Data.
        dim (str): Dimension name.
    
    Returns:
        bool: True if dim on object, False otherwise.
    """
    return dim in list(dx.dims)


def has_attr(obj, attr):
    """Test if obj has the attribute attr.

    Args:
        obj (xarray.DataArray or xarray.Dataset): xarray object.
        attr (str): Name of the attribute.

    Returns:
        bool: True if obj has attribute attr, False otherwise.
    """
    return attr in obj.attrs.keys()


def isolate_coordinate(obj, coordinate_name, drop=True):
    """Ensure that coordinate_name is the only coordinate on obj by selecting the first index of all other coordinates.

    Args:
        obj (xarray.DataArray or xarray.Dataset): xarray object.
        coordinate_name (str): Name of the coordinate to retain.
        drop (bool, optional): Drop the coordinate_name from the object. Defaults to True.
    """
    for coord in obj.coords.keys():
        if coord != coordinate_name:
            obj = obj.isel(**{coord: 0}, drop=drop)
    return obj


def extract_metadata(ds):
    """Extract metadata from an xarray dataset.

    Args:
        ds (str or xarray.Dataset): Path to (netcdf) or dataset.

    Returns:
        dict : Metadata dictionary.
    """
    if isinstance(ds, str):
        ds = xr.open_dataset(ds)

    metadata = dict(
        _global=dict(),
        variables=dict()
    )

    # Add global attributes
    for key, value in ds.attrs.items():
        metadata['_global'][key] = infer_dtype(value)

    # Add variable attributes (includes coordinates)
    for v in get_variables_and_coordinates(ds):

        metadata['variables'][v] = dict()

        for key, value in ds[v].attrs.items():
            metadata['variables'][v][key] = infer_dtype(value)

    return metadata


def load_metadata_json(filepath):
    """Load a metadata.json file.

    Args:
        filepath (str): Path to the json file.

    Returns:
        dict: Metadata dictionary.
    """
    metadata = json.loads(open(filepath, 'r').read())
    return metadata


def load_schema_json(filepath):
    """Load a schema.json file.

    Args:
        filepath (str): Path to the json file.

    Returns:
        dict : Schema dictionary.
    """
    return load_metadata_json(filepath)


def load_cf_standard_name_table(filepath):
    """Load and convert a CF standard name table XML into a metadata schema.

    Args:
        filepath (str): Path to the file.

    Returns:
        dict : Metadata dictionary (will be long).
    """
    xml = et.parse(filepath)
    root = xml.getroot()

    # Set up the schema information
    schema = dict(
        name='CF Standard Names Table',
        description='Specification converted from the CF standard names table',
        original_file=filepath.split('/')[-1],
        variables=dict()
    )

    # Header mappings
    header_mapping = dict(
        version_number='version',
        institution='contact',
        contact='contact_email',
        last_modified='created'
    )

    # Variable mappings
    child_mapping = dict(
        canonical_units='units',
        grib='grib',
        amip='amip',
        description='description'
    )

    # Iterate through the children
    for child in root:

        # Map the header information
        if child.tag not in ['entry', 'alias']:
            schema[header_mapping[child.tag]] = child.text

        # Variable information
        if child.tag == 'entry':

            # Get the name of the variable
            key = child.attrib['id']

            _schema = dict(
                standard_name={
                    'type': 'string',
                    'allowed': [key]
                }
            )

            # Iterate through the required metadata attributes
            for _child in child:

                _key = child_mapping[_child.tag]

                _schema[_key] = {
                    'type': 'string'
                }

                # Allow for nullable fields
                if _child.text:
                    _schema[_key]['allowed'] = [_child.text]

            # Add variable schema to the main schema
            schema['variables'][key] = _schema

        # Copy aliases
        if child.tag == 'alias':
            other = child.find('entry_id').text
            schema['variables'][child.attrib['id']] = schema['variables'][other]

    return schema


def save_schema(schema, filepath):
    """Save the schema to a json file at filepath.

    Args:
        schema (dict): Schema dictionary.
        filepath (str): Path to which to save the schema.
    """
    json.dump(schema, open(filepath, 'w'), indent=4)


def get_timestamp():
    """Generate a JSON-compliant timestamp.

    Returns:
        str : Timestamp.
    """
    return datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%SZ')


def load_cordex_csv(filepath, **kwargs):
    """Load a cordex csv (i.e. from CCAM) and convert it into a metadata schema.

    Note: There are no _global attributes.

    Args:
        filepath (str): Path to file
        **kwargs : Additional key/value pairs to add to the schema header (i.e. contact etc.)

    Returns:
        dict : Schema dictionary
    """
    # Load the CSV verbatim
    df = pd.read_csv(filepath)

    schema = dict(
        name='CORDEX Metadata Specification',
        description='Converted from CSV',
        original_file=filepath.split('/')[-1],
        created=get_timestamp(),
        **kwargs,
        _global=dict()
    )

    # Convert the lines into metadata specifications
    variables = dict()

    # Loop through variables
    for line in df.to_dict('records'):

        key = line['variable']
        variables[key] = dict()

        # Loop through attributes
        for attribute, value in line.items():

            # Skip variable, we already have it
            if attribute == 'variable':
                continue

            variables[key][attribute] = {
                'type': 'string',
            }

            # Check if there is an expected value
            if pd.isna(value) == False:
                variables[key][attribute]['allowed'] = [value]

    schema['variables'] = variables
    return schema


def load_package_data(slug, package_name='axiom', return_type='json'):
    """Load data from the installed Axiom package.

    Args:
        slug (str): Internal data path.
        package_name (str) : Name of the package. Defaults to 'axiom'.

    Returns:
        mixed : Dictionary of data (JSON only ATM).
    """
    _raw = pkgutil.get_data(package_name, slug)

    # Allow return of raw data.
    if return_type == 'json':
        return json.loads(_raw.decode('utf-8'))
    
    if return_type == 'config':
        parser = ConfigParser()
        parser.read_string(_raw.decode('utf-8'))
        return parser

    if return_type == 'raw':
        return _raw

    raise Exception('Unknown return type.')


def apply_schema(ds, schema):
    """Apply a metadata schema on a dataset.

    Args:
        ds (xarray.Dataset): Dataset.
        schema (dict): Axiom schema dictionary.

    Returns:
        xarray.Dataset : Dataset with schema-defined metadata applied.
    """
    # Apply the global metadata, if available
    for key, _schema in schema['_global'].items():
        if 'allowed' in _schema.keys():
            ds.attrs[key] = _schema['allowed'][0]

    # Loop through each variable on the dataset
    for variable in list(ds.data_vars.keys()) + list(ds.coords.keys()):

        # If the variable is define
        if variable in schema['variables'].keys():

            # Extract the schema for the variable itself
            var_schema = schema['variables'][variable]

            # Loop through attributes
            for key, _schema in var_schema.items():

                # Apply if there is an expected value
                if 'allowed' in _schema.keys():
                    ds[variable].attrs[key] = _schema['allowed'][0]

    # Return the updated schema
    return ds

def _diff_metadata(meta_a, meta_b, ignore_matches=True):

    # Dictionaries are equal
    if meta_a == meta_b:
        diff = {key: (None, None) for key in meta_a.keys()}
        return diff

    # Otherwise we need to parse them
    diff = dict()

    # Check attributes from A
    parsed_keys = list()
    for key, value1 in meta_a.items():

        # Missing from B
        if key not in meta_b.keys():
            diff[key] = (value1, None)

        # Same value
        elif value1 == meta_b[key] and ignore_matches == False:
            diff[key] = (None, None)

        # Different value
        elif value1 != meta_b[key]:
            diff[key] = (value1, meta_b[key])

        # Mark as parsed for meta_b
        parsed_keys.append(key)

    # Check attributes of B
    for key, value2 in meta_b.items():

        # Skip already parsed
        if key in parsed_keys:
            continue

        # This will be the only test, the inverse has already been checked
        if key not in meta_a.keys():
            diff[key] = (None, value2)

    return diff


def diff_metadata(meta_a, meta_b, ignore_matches=True):
    """Difference the metadata between two metadata dictionaries.

    Differences are encoded with values as tuples, where:
    - (None, None) = Attributes match.
    - (None, value2) = Attribute is missing from meta_a.
    - (value1, None) = Attribute is missing from meta_b.
    - (value1, value2) = Differing values between meta_a and meta_b.

    Args:
        meta_a (dict): Metadata dictionary of the form from extract_metadata.
        meta_b (dict): Metadata dictionary of the form from extract_metadata.

    Returns:
        dict : Dictionary of differences.
    """

    # Do the global comparison
    diff = dict(
        _global=_diff_metadata(
            meta_a['_global'],
            meta_b['_global'],
            ignore_matches=ignore_matches
        ),
        variables=dict()
    )

    # Do the comparison of all the variables that the dicts share
    for variable in meta_a['variables'].keys():

        if variable in meta_b.keys():
            diff['variables'][variable] = _diff_metadata(
                meta_a[variable],
                meta_b[variable],
                ignore_matches=ignore_matches
            )

    return diff


def infer_dtype(value):
    """Infer the data type of the value passed.

    Args:
        value (unknown): Value.

    Raises:
        ValueError : When the type can't be inferred.
    """

    if isinstance(value, bool):
        return value

    # for dtype in [float, int, bool, str]:
    for dtype in [float, int, str]:

        try:
            return dtype(value)
        except ValueError:
            pass

    raise ValueError('Unable to infer type.')


def auto_glob(path):
    """Shorthand for sorted(glob.glob(path))

    Args:
        mask (str): Globabble path
    
    Returns:
        list : List of paths that match the globbable part.    
    """
    if not isinstance(path, list):
        return sorted(glob.glob(path))
    
    return path


def touch(filepath):
    """Thin wrapper for touching files.

    Args:
        filepath (str): Path.
    """
    Path(filepath).touch()


def get_lock_filepath(filepath):
    """Get a lock file path for filepath.

    Args:
        filepath (str): Path.

    Returns:
        str: Lock file path.
    """
    return f'{filepath}.lock'


def lock(filepath):
    """Place a lock on a filepath.

    Args:
        filepath (str): Path to the file.
    """
    touch(get_lock_filepath(filepath))


def unlock(filepath):
    """Remove a lock on a filepath.

    Args:
        filepath (str): Path to the file.
    """
    os.remove(get_lock_filepath(filepath))


def is_locked(filepath):
    """Check if a file is locked.

    Args:
        filepath (str): Path to the file.
    """
    return os.path.isfile(get_lock_filepath(filepath))


class ListAwareConfigParser(ConfigParser):

    def __init__(self):
        super().__init__(interpolation=ExtendedInterpolation())

    def getlist(self, section, key, delimiter='\n', cast=str):
        """Get an option out of the configuration and automatically split into a list.

        Args:
            section (str): Section name.
            key (str): Option key.
            delimiter (str, optional): Delimiter. Defaults to '\n'.
            cast (callable): Typecast on the fly to this type.

        Returns:
            list: Configuration option as a list.
        """
        values = self.get(section, key).split(delimiter)
        values = list(filter(bool, values))
        return [cast(v) for v in values]
    
    def section2dict(self, section, detect_dtypes=True):

        _dict = dict()
        for option in self._sections[section].keys():
            if detect_dtypes:
                _dict[option] = infer_dtype(self.get(section, option))
            else:
                _dict[option] = self.get(section, option)
        
        return _dict


def load_user_config(name):
    """Load a configuration file from the user's $HOME/.axiom directory.

    Args:
        name (str): Name of the configuration file.
    
    Returns:
        configparser.Config : Configuration.
    """
    # Build a path to the user's home directory for that configuration file.
    config_path = os.path.join(
        str(Path.home()),
        '.axiom',
        f'{name}.ini'
    )

    if os.path.isfile(config_path) is False:
        raise FileNotFoundError(f'No configuration file found at {config_path}')

    config = ListAwareConfigParser()
    config.read(config_path)
    return config


def pluralise(obj):
    """Automatically convert obj into a list object.

    For use when an object can be defined singularly or plurally.

    Args:
        obj (object): Object.
    """
    if not isinstance(obj, list):
        obj = [obj]
    
    return obj


class Timer:
    """Basic timer class."""

    def __init__(self):
        self._start_time = None
    
    def start(self):
        """Start the timer."""
        self._start_time = time.perf_counter()
    
    def stop(self):
        """Stop the timer.
        
        Returns:
            float : Elapsed time in seconds"""
        elapsed_time = time.perf_counter() - self._start_time
        self._start_time = None
        return elapsed_time


def shell(cmd, shell=True, check=True, capture_output=True, **kwargs):
    """Execute a shell command.

    Args:
        cmd (str): Command
        **kwargs : Passed to subprocess.run

    Returns:
        [type]: [description]
    """
    return sp.run(cmd, shell=shell, check=check, capture_output=capture_output, **kwargs)


def batch_split(iterable, n_batches):
    """Split iterable into n_batches.
    
    Args:
        iterable (iterable) : Iterable object to split.
        n_batches (int) : Number of batches.
    
    Returns:
        list : List of iterables.
    """
    if n_batches == 1:
        return [iterable]

    return np.array_split(iterable, n_batches)


def conditional_rename(ds, **kwargs):
    """Conditionally rename an object on a dataset, if it exists.

    Args:
        ds (xarray.Dataset): Dataset.
        **kwargs : Key/value pairs of old=new names.

    Returns:
        xarray.Dataset : Dataset with renamed variables/coords.
    """
    # Start with an empty mapping.
    renames = dict()

    # If the old exists, then add it to the mapping.
    for old, new in kwargs.items():
        if in_ds(ds, old):
            renames[old] = new
    
    # If there are any mappings to apply, do them now
    if renames:
        return ds.rename(**renames)

    # Return the original
    return ds


def interpolate_template(template, **kwargs):
    """Interpolate onto a template string.
    
    Args:
        string (str): Template string.
        **kwargs : Key/value pairs to interpolate.
    
    Returns:
        str : Interpolated string
    """
    _template = Environment(loader=BaseLoader()).from_string(template)
    return _template.render(**kwargs)


def get_installed_data_root():
    """Get the root directory for installed data.
    
    Returns:
        str : Root directory.
    """
    return os.path.join(importlib.resources.files('axiom'), 'data')


def get_user_data_root():
    """Get the root directory for user data.
    
    Returns:
        str : Root directory.
    """
    return os.path.join(Path.home(), '.axiom')


def load_package_json(path):
    """Load package JSON data from the Axiom installation directory or from the user's home directory.

    Args:
        path (str): Path to the data.
    
    Returns:
        dict : Data.
    """
    # Try to load from the user's home directory first.
    user_filepath = os.path.join(get_user_data_root(), path)
    installed_filepath = os.path.join(get_installed_data_root(), path)

    for _path in [user_filepath, installed_filepath]:
        if os.path.isfile(_path):
            return json.load(open(_path, 'r'))
    
    raise FileNotFoundError(f'No installed or user data found at {path}')


"""Code to test the supervisor class."""
import pytest
import time
from axiom.supervisor import Supervisor

def test_supervisor_fail():
    """Test the supervisor class will correctly detect a long-running job."""

    # Test that a long running job raises an exception.
    with pytest.raises(TimeoutError):
        with Supervisor(seconds=3, error_msg='Job took too long.'):
            time.sleep(5)
    
def test_supervisor_pass():
    """Test the supervisor class will not detect a jobs that completes in time."""
    # Test a job that should complete in time.
    with Supervisor(seconds=5, error_msg='Job that should work.'):
        time.sleep(2)
        assert True


"""Decorators."""
import functools
import xarray as xr


def singleton(cls):
    """Define a class as a singleton.

    Returns:
        object: Singleton
    """

    instance = [None]
    
    @functools.wraps(cls)
    def _singleton(*args, **kwargs):
        if instance[0] is None:
            instance[0] = cls(*args, **kwargs)
        return instance[0]
    
    return _singleton


def metadata(**meta):
    """Apply metadata to the xarray outputs of a function.

    Usage Example:

    @metadata(units='K')
    def get_temp(ds):
        return ds.T

    """
    def _metadata(func):
        @functools.wraps(func)
        def __metadata(*args, **kwargs):
            result = func(*args, **kwargs)

            # Allow name to be set on DataArrays, special edge case.
            if 'name' in meta.keys() and isinstance(result, xr.DataArray):
                result.name = meta.pop('name')  

            # Allow the user to completely replace any existing metadata
            if 'replace' in meta.keys() and meta['replace'] is True:
                meta.pop('replace')
                result.attrs = meta
            else:
                result.attrs.update(meta)
            
            return result

        return __metadata
    
    return _metadata
    